import os
import logging
import math
from time import time
import threading

from PyQt5 import QtWidgets, uic, QtCore, QtGui
from matplotlib import animation, cm
from matplotlib.colors import Normalize
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas
)
from matplotlib.animation import ArtistAnimation
import numpy as np

import plots
from utilities import xycoord


UI_FILE = os.path.join(os.path.dirname(__file__), 'qt_frame_window.ui')
log = logging.getLogger(__name__)


class FrameAnimation(animation.ArtistAnimation):  # pragma: no cover
    """Performs the animation for scrolling through frames arbitarily.
    """
    
    def __init__(self, fig, artists, *args, **kwargs):
        self.fig = fig
        self.artists = artists
        log.debug("Creating animation in thread %d", threading.get_ident())
        # Call parent init methods
        animation.ArtistAnimation.__init__(self, *args, fig=fig,
                                           artists=artists, **kwargs)
    
    def _step(self, current_idx):
        log.debug("Animating to frame {}".format(current_idx))
        artists = self.artists[current_idx]
        self._draw_next_frame(artists, self._blit)
        return True
    
    def stop(self):
        log.debug("Stopping animation in thread %d", threading.get_ident())
        return self._stop()


class FrameChangeSource(QtCore.QObject):
    _is_running = False
    callbacks = []
    
    def __init__(self, view, *args, **kwargs):
        log.debug('creating source')
        self.callbacks = [] # To avoid a Borg callback list
        self.view = view
        super().__init__(*args, **kwargs)
    
    def add_callback(self, func, *args, **kwargs):
        self.callbacks.append((func, args, kwargs))
    
    def remove_callback(self, func, *args, **kwargs):
        if args or kwargs:
            self.callbacks.remove((func, args, kwargs))
        else:
            funcs = [c[0] for c in self.callbacks]
            if func in funcs:
                self.callbacks.pop(funcs.index(func))
    
    def start(self):
        if not self._is_running:
            log.debug("Starting FrameChangeSource in thread %d", threading.get_ident())
            # Listen to the frame adjustment signals
            self.view.presenter_frame_changed.connect(self._on_change)
            self._is_running = True
    
    def stop(self):
        if self._is_running:
            log.debug("Stopping FrameChangeSource in thread %d", threading.get_ident())
            self.view.presenter_frame_changed.disconnect(self._on_change)
            self._is_running = False
    
    def _on_change(self, new_idx):
        for func, args, kwargs in self.callbacks:
            func(new_idx, *args, **kwargs)


class QtFrameView(QtCore.QObject):  # pragma: no cover
    window = None
    ui = None
    fig = None
    spectrum_ax = None
    hist_ax = None
    hist_cb = None
    edge_ax = None
    _frame_animation = None
    
    # Signals
    new_hdf_file_requested = QtCore.pyqtSignal(bool)
    new_hdf_group_requested = QtCore.pyqtSignal(QtWidgets.QTreeWidgetItem, QtWidgets.QTreeWidgetItem,
                                                arguments=('new_item', 'old_item',))
    frame_slider_moved = QtCore.pyqtSignal(int)
    presenter_frame_changed = QtCore.pyqtSignal(int)
    new_vrange_requested = QtCore.pyqtSignal(float, float)
    reset_vrange_requested = QtCore.pyqtSignal()
    new_timestep_requested = QtCore.pyqtSignal(int)
    new_cmap_requested = QtCore.pyqtSignal(str)
    new_component_requested = QtCore.pyqtSignal(str)
    play_button_clicked = QtCore.pyqtSignal(bool)
    forward_button_clicked = QtCore.pyqtSignal()
    back_button_clicked = QtCore.pyqtSignal()
    first_button_clicked = QtCore.pyqtSignal()
    last_button_clicked = QtCore.pyqtSignal()
    play_speed_requested = QtCore.pyqtSignal(int)
    figure_hovered = QtCore.pyqtSignal(object)
    
    def build_hdf_tree(self, tree):
        icons = {
            'metadata': QtGui.QIcon.fromTheme('x-office-spreadsheet'),
            'frameset': QtGui.QIcon.fromTheme('emblem-photos'),
            'map': QtGui.QIcon.fromTheme('image-x-generic'),
        }
       
        # Recursive helper function for building the tree
        def tree_item(data, parent=None):
            item = QtWidgets.QTreeWidgetItem()
            item.setText(0, data['name'])
            item.setText(2, data['path'])
            if data['context'] is not None:
                item.setText(1, data['context'])
                item.setIcon(1, icons[data['context']])
            # Now recursively add children
            for child in data['children']:
                new_item = tree_item(child)
                child_item = tree_item(child)
                ret = item.addChild(child_item)
            return item
        
        # Start building the tree
        for child in tree:
            # Recursively create the tree item and its children
            new_root_item = tree_item(child)
            # Now add the new root to the tree
            self.add_hdf_tree_item(new_root_item)
            # Flag for expanding the active tree auotmatically
            child_name = child['path'].split('/')[1]
    
    def add_hdf_tree_item(self, item):
        self.ui.hdfTree.addTopLevelItem(item)
    
    def select_active_hdf_item(self, item):
        self.ui.hdfTree.setCurrentItem(item)
    
    def setup_ui(self):
        # Load the Qt Designer .ui file
        Ui_FrameWindow, QMainWindow = uic.loadUiType(UI_FILE)
        log.debug("Built frame window using uic") 
        # Create the UI elements
        self.window = QtWidgets.QMainWindow()
        self.ui = Ui_FrameWindow()
        self.ui.setupUi(self.window)
        # Add labels to the HDF Tree widget
        self.ui.hdfTree.setHeaderLabels(['Name', 'Type'])
        header = self.ui.hdfTree.header()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        header.setStretchLastSection(False)
        # Connect the signal for drawing the frames, histogram, etc.
        self.create_canvas()
        # Create animations from the collected artists
        self.source = FrameChangeSource(view=self)
        # Connect UI signals to view signals
        self.ui.sldFrameSlider.valueChanged.connect(self.frame_slider_moved)
        self.ui.btnApplyLimits.clicked.connect(self.request_new_vrange)
        self.ui.btnResetLimits.clicked.connect(self.reset_vrange_requested)
        self.ui.cmbTimestep.currentIndexChanged.connect(self.new_timestep_requested)
        self.ui.cmbCmap.currentTextChanged.connect(self.new_cmap_requested)
        self.ui.cmbComponent.currentTextChanged.connect(self.new_component_requested)
        self.ui.btnPlay.toggled.connect(self.play_button_clicked)
        self.ui.btnForward.clicked.connect(self.forward_button_clicked)
        self.ui.btnBack.clicked.connect(self.back_button_clicked)
        self.ui.btnFirst.clicked.connect(self.first_button_clicked)
        self.ui.btnLast.clicked.connect(self.last_button_clicked)
        self.ui.sldPlaySpeed.valueChanged.connect(self.play_speed_requested)
        self.ui.hdfTree.currentItemChanged.connect(self.new_hdf_group_requested)
        self.ui.actionOpen.triggered.connect(self.new_hdf_file_requested)
        # Emit a signal when mouse hovers over the figure
        def hover_frame(event):
            if event.inaxes is self.img_ax:
                xy = xycoord(x=event.xdata, y=event.ydata)
            else:
                xy = None
            self.figure_hovered.emit(xy)
        self.fig.canvas.mpl_connect('motion_notify_event', hover_frame)
        # Launch the actual window itself
        self.window.show()
    
    def request_new_vrange(self, bool):
        vmin = self.ui.spnVMin.value()
        vmax = self.ui.spnVMax.value()
        self.new_vrange_requested.emit(vmin, vmax)
    
    def create_status_bar(self):
        self.status_layout = QtWidgets.QVboxLayout()
        self.ui.statusbar.layout().addItem(QtWidgets.QSpacerItem(100, 20))
        # Indicator for the current frame shape
        self.ui.statusbar.addWidget(QtWidgets.QLabel("Shape:"))
        self.lblShape = QtWidgets.QLabel()
        self.lblShape.setMinimumWidth(120)
        self.ui.statusbar.addWidget(self.lblShape)
        # Indicator for the current index
        self.ui.statusbar.addWidget(QtWidgets.QLabel("Frame:"))
        self.lblIndex = QtWidgets.QLabel()
        self.lblIndex.setMinimumWidth(30)
        self.ui.statusbar.addWidget(self.lblIndex)
        # Indicator for the current energy
        self.ui.statusbar.addWidget(QtWidgets.QLabel("Energy:"))
        self.lblEnergy = QtWidgets.QLabel()
        self.lblEnergy.setMinimumWidth(100)
        self.ui.statusbar.addWidget(self.lblEnergy)
        # Indicator for the current cursor position
        self.ui.statusbar.addWidget(QtWidgets.QLabel("Cursor x,y:"))
        self.lblCursor = QtWidgets.QLabel()
        self.lblCursor.setMinimumWidth(120)
        self.ui.statusbar.addWidget(self.lblCursor)
        # Indicator for the current pixel position
        self.ui.statusbar.addWidget(QtWidgets.QLabel("Pixel:"))
        self.lblPixel = QtWidgets.QLabel()
        self.lblPixel.setMinimumWidth(100)
        self.ui.statusbar.addWidget(self.lblPixel)
        # Indicator for the current cursor value
        self.ui.statusbar.addWidget(QtWidgets.QLabel("Value:"))
        self.lblValue = QtWidgets.QLabel()
        self.lblValue.setMinimumWidth(100)
        self.ui.statusbar.addWidget(self.lblValue)
    
    def create_canvas(self):
        # Add the canvas to the UI
        self.fig = Figure()
        canvas = FigureCanvas(self.fig)
        self.ui.mainLayout.addWidget(canvas)
        # Add subplots
        gridspec = GridSpec(2, 4)
        self.img_ax = self.fig.add_subplot(
            gridspec.new_subplotspec((0, 0), rowspan=2, colspan=2)
        )
        self.spectrum_ax = self.fig.add_subplot(
            gridspec.new_subplotspec((0, 2), rowspan=1, colspan=2)
        )
        self.hist_ax = self.fig.add_subplot(
            gridspec.new_subplotspec((1, 2), rowspan=1, colspan=1)
        )
        self.edge_ax = self.fig.add_subplot(
            gridspec.new_subplotspec((1, 3), rowspan=1, colspan=1)
        )
        # Create the colorbar on the histogram axes
        self.cbar = plots.draw_histogram_colorbar(ax=self.hist_ax,
                                                  cmap="viridis",
                                                  norm=Normalize(0, 1))
        self.cbar.ax.set_xlabel("Intensity")
        # Adjust the margins
        self.fig.tight_layout(pad=0)
        self.fig.canvas.draw_idle()
    
    def connect_presenter(self, presenter):
        presenter.frame_data_changed.connect(self.animate_frames)
        presenter.frame_data_changed.connect(self.set_status_metadata)
        presenter.frame_data_changed.connect(self.draw_histogram)
        presenter.frame_data_cleared.connect(self.clear_axes)
        presenter.mean_spectrum_changed.connect(self.draw_spectrum)
        presenter.active_frame_changed.connect(self.presenter_frame_changed.emit)
        presenter.active_frame_changed.connect(self.move_slider)
        presenter.frame_vrange_changed.connect(self.change_vrange)
        presenter.active_energy_changed.connect(self.set_status_energy)
        presenter.frame_hover_changed.connect(self.update_figure_pixel)
        # Respond to changes in available combo box options
        presenter.cmap_list_changed.connect(self.set_cmap_list)
        presenter.component_list_changed.connect(self.set_component_list)
        presenter.timestep_list_changed.connect(self.set_timestep_list)
        # Respond to a new HDF data tree
        presenter.hdf_tree_changed.connect(self.build_hdf_tree)
        presenter.hdf_path_changed.connect(self.set_window_title)
        # Prepare the UI after the application has been launched
        presenter.app_ready.connect(self.setup_ui)
    
    def connect_signals(self, presenter):
        pass
        # Connect internal signals and slots
        # Opening a new HDF file
        # Connect a handler for when the user hovers over the frame
   
    def open_hdf_file(self):
        # Ask the user for an HDF file to open
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self.window, "QFileDialog.getOpenFileName()", "",
            "All Files (*);;HDF5 Files (*.h5 *.hdf *.hdf5)",
            'HDF5 Files (*.h5 *.hdf *.hdf5)')
        self.file_opened.emit(filename)
        return filename
    
    def frame_controls(self):
        """Gives a list of all the UI buttons that are associated with
        changing the currently active frame."""
        widgets = [self.ui.btnFirst, self.ui.btnBack, self.ui.btnPlay,
                   self.ui.btnForward, self.ui.btnLast,
                   self.ui.sldPlaySpeed, self.ui.sldFrameSlider]
        return widgets
    
    def plotting_controls(self):
        """Gives a list of all the UI elements that are associated with
        changing how the frameset is plotted."""
        widgets = [self.ui.cmbTimestep, self.ui.cmbComponent, self.ui.cmbCmap,
                   self.ui.spnVMin, self.ui.spnVMax,
                   self.ui.btnApplyLimits, self.ui.btnResetLimits]
        return widgets
    
    def move_slider(self, val):
        self.ui.sldFrameSlider.setValue(val)
        self.ui.lblIndex.setText(str(val))
    
    def set_busy_mode(self, status):
        """Either enable or disable the busy mode.
        
        Changes the mouse pointer and disables/enables controls.
        
        Parameters
        ----------
        status : bool
          If truthy, the UI is disabled, otherwise the UI is
          re-enabled.
        
        """
        # Disable all the controls so the user can't interact
        controls = self.frame_controls() + self.plotting_controls()
        for ctrl in controls:
            ctrl.setDisabled(status)
        # Use the busy cursor for the window
        if status:
            self.window.setCursor(QtCore.Qt.WaitCursor)
        else:
            self.window.unsetCursor()
        # Print a message informing the user of what we're doing
        if status:
            self.ui.statusbar.showMessage("Drawing...")
        else:
            self.ui.statusbar.clearMessage()
    
    def animate_frames(self, frames, energies, norm, cmap, extent):
        self.set_busy_mode(True)
        # Disconnect the old animation first
        if self._frame_animation:
            log.debug("Disconnecting old frame animations")
            self._frame_animation.stop()
            del self._frame_animation
        # Prepare the axes for new plotting
        artists = []
        self.img_ax.clear()
        # Create the new artists
        start = time()
        for energy, frame in zip(energies, frames):
            im_artist = plots.plot_txm_map(frame, norm=norm,
                                           cmap=cmap, ax=self.img_ax,
                                           extent=extent)
            line_artist = self.spectrum_ax.axvline(energy,
                                                   linestyle=":",
                                                   animated=True,
                                                   color="gray")
            edge_artist = self.edge_ax.axvline(energy,
                                               linestyle=":",
                                               animated=True,
                                               color="gray")
            artists.append([im_artist, line_artist, edge_artist])
        # Create animations from the collected artists
        self._frame_animation = FrameAnimation(
            fig=self.fig,
            artists=artists,
            event_source=self.source,
            blit=True)
        # Redraw the canvas (this part is slow)
        self.fig.canvas.draw_idle()
        # User feedback and logging
        log.debug("Artist creation took %d sec", time() - start)
    
    def draw_spectrum(self, spectrum, fitted_spectrum, norm, cmap, edge_range=None):
        self.spectrum_ax.clear()
        plots.plot_spectrum(spectrum=spectrum.values,
                            energies=spectrum.index,
                            norm=norm,
                            cmap=cmap,
                            color="y",
                            ax=self.spectrum_ax)
        # Now plot a zoomed in version on the edge itself
        self.edge_ax.clear()
        plots.plot_spectrum(spectrum=spectrum.values,
                                  energies=spectrum.index,
                                  norm=norm,
                                  cmap=cmap,
                                  color="y",
                                  ax=self.edge_ax)
        if edge_range is not None:
            self.edge_ax.set_xlim(*edge_range)
    
    def draw_histogram(self, data, energies, norm, cmap):
        log.debug("Plotting frame histogram")
        # Update the histogram
        self.hist_ax.clear()
        plots.plot_txm_histogram(data=data, ax=self.hist_ax,
                                 cmap=cmap, norm=norm, add_cbar=False)
        self.hist_ax.xaxis.set_visible(False)
        # Update the colorbar
        mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
        self.cbar.on_mappable_changed(mappable)
        self.cbar.set_ticks(np.linspace(norm.vmin, norm.vmax, num=5))
        self.set_busy_mode(False)
    
    def clear_axes(self):
        self.img_ax.clear()
        self.spectrum_ax.clear()
        self.hist_ax.clear()
        self.edge_ax.clear()
        self.fig.canvas.draw()
    
    def set_slider_max(self, val):
        self.ui.sldFrameSlider.setRange(0, val)
    
    def update_figure_pixel(self, xy, xy_unit, px, value):
        # Set the xy label text
        if xy is None:
            xy_s = '(---, ---) {}'.format(xy_unit)
        else:
            xy_s = "({x:.3f}, {y:.3f}) {unit}".format(x=xy.x, y=xy.y, unit=xy_unit)
        self.ui.lblCursor.setText(xy_s)
        # Set the pixel label text
        if px is None:
            px_s = '[--, --]'
        else:
            px_s = "[{}, {}]".format(px[0], px[1])
        self.ui.lblPixel.setText(px_s)
        # Set the actual label value
        if value is None:
            val_s = '---'
        else:
            val_s = str(value)
        self.ui.lblValue.setText(val_s)
    
    def change_vrange(self, vmin, vmax, step, decimals):
        """Change the range of frame map colors.
        
        Does not affect the rendered frames."""
        self.ui.spnVMin.setValue(vmin)
        self.ui.spnVMax.setValue(vmax)
        self.ui.spnVMin.setSingleStep(step)
        self.ui.spnVMax.setSingleStep(step)
        self.ui.spnVMin.setDecimals(decimals)
        self.ui.spnVMax.setDecimals(decimals)
        # Set the min and max of the spin boxes so they don't cross
        self.ui.spnVMin.setMaximum(vmax - step)
        self.ui.spnVMax.setMinimum(vmin + step)
    
    def set_cmap_list(self, cmap_list):
        self.ui.cmbCmap.clear()
        self.ui.cmbCmap.addItems(cmap_list)
        self.ui.cmbCmap.setCurrentText('copper')
    
    def set_component_list(self, component_list):
        self.ui.cmbComponent.clear()
        self.ui.cmbComponent.addItems(component_list)
    
    def set_timestep_list(self, timestep_list):
        self.ui.cmbTimestep.clear()
        self.ui.cmbTimestep.addItems(timestep_list)
    
    def set_timestep(self, idx):
        self.ui.cmbTimestep.setCurrentIndex(idx)
    
    def show_status_message(self, message):
        self.ui.statusbar.showMessage(message)
    
    def set_status_metadata(self, frames, energies, norm, cmap, extent):
        self.ui.lblShape.setText(str(frames.shape))
        self.ui.lblDtype.setText(str(frames.dtype))
        # Set the slider maximum
        self.ui.sldFrameSlider.setMaximum(frames.shape[0]-1)
    
    def set_status_index(self, msg):
        self.ui.lblIndex.setText(msg)
    
    def set_status_energy(self, energy):
        s = "{:.2f} eV".format(energy)
        self.ui.lblEnergy.setText(s)
    
    def set_status_unit(self, msg):
        self.ui.lblUnit.setText(msg)
    
    def set_window_title(self, hdf_path):
        title = "Xanespy Frameset ({})".format(hdf_path)
        self.window.setWindowTitle(title)



# -*- coding: utf-8 -*-
#
# Copyright © 2016 Mark Wolf
#
# This file is part of Xanespy.
#
# Xanespy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Xanespy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Xanespy.  If not, see <http://www.gnu.org/licenses/>.

"""Define classes for more fine-grained control over exception
handling."""


#########################
# X-ray microscopy errors
#########################
class DataNotFoundError(FileNotFoundError):
    """Expected a directory containing data but found none."""
    pass


class FrameFileNotFound(IOError):
    """Expected to load a TXM frame file but it doesn't exist."""
    pass


class GroupKeyError(KeyError):
    """Tried to load or create an HDF group but failed. Examples include:
    the group doesn't exist, is ambiguous or already exists when being
    created.
    """
    pass


class FrameSourceError(KeyError):
    """The frame-source attribute is not present."""
    pass


class DataFormatError(RuntimeError):
    """The raw data are arranged in a way that the importers or TXM classes do
    not understand.
    """
    pass


class HDFScopeError(ValueError):
    """Tried to pass an HDF scope that is not recognized."""
    pass


class FileExistsError(IOError):
    """Tried to import a TXM frameset but the corresponding HDF file
    already exists."""
    pass


class CreateGroupError(ValueError):
    """Tried to import a TXM frameset into a group but the corresponding
    HDF group already exists or is otherwise inaccessible.
    """
    pass


class FilenameParseError(ValueError):
    """The parameters in the filename do not match the naming scheme
    associated with this flavor."""
    pass


class DatasetExistsError(RuntimeError):
    """Trying to save a new dataset but one already exists with the given
    path."""
    pass


class NoParticleError(Exception):
    pass


class RefinementError(RuntimeError):
    pass


class XanesMathError(RuntimeError):
    pass



import os
import logging

from PyQt5 import QtWidgets, uic, QtCore
from matplotlib import cm
from matplotlib.colors import Normalize
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas
)

import plots

log = logging.getLogger(__name__)


UI_FILE = os.path.join(os.path.dirname(__file__), 'qt_map_window.ui')

# A list of colors that give good contrast with each colormap
CROSSHAIR_COLORS = {
    'plasma': '#20f986',  # wintergreen
    'viridis': '#fd5956',  # grapefruit
    'inferno': '#20f987',  # wintergreen
    'magma': '#20f986',  # wintergreen
    'gray': 'red',
    'bone': 'red',
    'copper': 'green',
    'autumn': 'blue',
    'spring': 'blue',
    'summer': 'red',
    'winter': 'red',
    'spectral': 'black',
    'rainbow': 'black',
    'gnuplot': 'green',
}


class QtMapView(QtCore.QObject):
    """A Qt view for a frameset map. It should be controlled by a
    presenter.
    
    Attributes
    ----------
    cmap_changed : signal
      Fires when the user changes the colormap via the UI
    limits_applied : signal
      Fires when the user requests that data be redrawn with new limits.
    limits_reset : signal
      Fires when the user asks that the norm limits be reset to the data.
    
    """
    window = None
    ui = None
    fig = None
    crosshairs = None
    latest_cmap = "plasma"
    
    # Signals
    cmap_changed = QtCore.pyqtSignal('QString')
    component_changed = QtCore.pyqtSignal('QString')
    edge_mask_toggled = QtCore.pyqtSignal(bool)
    spectrum_fit_toggled = QtCore.pyqtSignal(bool)
    map_vmin_changed = QtCore.pyqtSignal(float)
    map_vmax_changed = QtCore.pyqtSignal(float)
    limits_applied = QtCore.pyqtSignal()
    limits_reset = QtCore.pyqtSignal()
    map_hovered = QtCore.pyqtSignal(object, object)
    map_clicked = QtCore.pyqtSignal(object, object)
    map_moved = QtCore.pyqtSignal(int, int)
    
    def setup_ui(self):  # pragma: no cover
        Ui_FrameWindow, QMainWindow = uic.loadUiType(UI_FILE)
        log.debug("Built map window using uic")
        # Create the UI elements
        self.window = QtWidgets.QMainWindow()
        self.ui = Ui_FrameWindow()
        self.ui.setupUi(self.window)
        # Create the matplotlib elements
        self.create_canvas()
        # Connect the UI signals to the view signals
        self.ui.cmbCmap.currentTextChanged.connect(self.cmap_changed)
        self.ui.cmbComponent.currentTextChanged.connect(self.component_changed)
        self.ui.chkEdgeMask.toggled.connect(self.edge_mask_toggled)
        self.ui.chkFitSpectrum.toggled.connect(self.spectrum_fit_toggled)
        self.ui.spnVMin.valueChanged.connect(self.map_vmin_changed)
        self.ui.spnVMax.valueChanged.connect(self.map_vmax_changed)
        self.ui.btnApplyLimits.clicked.connect(self.limits_applied)
        self.ui.btnResetLimits.clicked.connect(self.limits_reset)
        self.fig.canvas.mpl_connect('motion_notify_event', self.mouse_in_canvas)
        self.fig.canvas.mpl_connect('button_release_event', self.mouse_clicked_canvas)
        self.fig.canvas.mpl_connect('key_press_event', self.keyboard_nav)
    
    def keyboard_nav(self, event):
        # Holds descriptions of how much to move for each key in pixels
        moves = {
            'up': (-1, 0),
            'w': (-1, 0),
            'down': (1, 0),
            's': (1, 0),
            'left': (0, -1),
            'a': (0, -1),
            'right': (0, 1),
            'd': (0, 1),
            'ctrl+up': (-10, 0),
            'ctrl+w': (-10, 0),
            'ctrl+down': (10, 0),
            'ctrl+s': (10, 0),
            'ctrl+left': (0, -10),
            'ctrl+a': (0, -10),
            'ctrl+right': (0, 10),
            'ctrl+d': (0, 10),
        }
        if event.key in moves.keys():
            self.map_moved.emit(*moves[event.key])
    
    def mouse_in_canvas(self, mouse_event):
        if mouse_event.inaxes is self.map_ax:
            self.map_hovered.emit(mouse_event.xdata, mouse_event.ydata)
        else:
            self.map_hovered.emit(None, None)
    
    def mouse_clicked_canvas(self, mouse_event):
        if mouse_event.inaxes is self.map_ax:
            log.debug("Clicked map: x=%f, y=%f", mouse_event.xdata, mouse_event.ydata)
            self.map_clicked.emit(mouse_event.xdata, mouse_event.ydata)
        else:
            self.map_clicked.emit(None, None)
    
    def update_cursor_labels(self, xy, pixel, value):
        xy_s, pixel_s, value_s = "", "", ""
        if xy is not None:
            xy_s = "{x:.2f}, {y:.2f}".format(x=xy.x, y=xy.y)
        if pixel is not None:
            pixel_s = "[{row}, {col}]".format(row=pixel.vertical, col=pixel.horizontal)
        if value is not None:
            value_s = "{}".format(value)
        self.ui.lblCursorXY.setText(xy_s)
        self.ui.lblCursorPixel.setText(pixel_s)
        self.ui.lblCursorValue.setText(value_s)
    
    def update_crosshair_labels(self, xy, pixel, value):
        xy_s, pixel_s, value_s = "", "", ""
        if xy is not None:
            xy_s = "{x:.2f}, {y:.2f}".format(x=xy.x, y=xy.y)
        if pixel is not None:
            pixel_s = "[{row}, {col}]".format(row=pixel.vertical, col=pixel.horizontal)
        if value is not None:
            value_s = "{}".format(value)
        self.ui.lblCrosshairsXY.setText(xy_s)
        self.ui.lblCrosshairsPixel.setText(pixel_s)
        self.ui.lblCrosshairsValue.setText(value_s)
    
    def redraw_crosshairs(self, xy):  # pragma: no cover
        """Draw a set of crosshairs on the map at location given by `xy`."""
        # Remove the old crosshairs first
        if self.crosshairs is not None:
            for artist in self.crosshairs:
                artist.remove()
            self.crosshairs = None
        # Plot new crosshairs on the map axes
        if xy is not None:
            log.debug("Drawing crosshairs at {}".format(xy))
            color = CROSSHAIR_COLORS[self._latest_cmap]
            xartist = self.map_ax.axvline(xy.x, linestyle="--", color=color, linewidth=1)
            yartist = self.map_ax.axhline(xy.y, linestyle="--", color=color, linewidth=1)
            self.crosshairs = [xartist, yartist]
        
    def show(self):  # pragma: no cover
        self.window.show()
    
    def hide(self):  # pragma: no cover
        if hasattr(self.window, 'hide'):
            self.window.hide()
    
    def redraw_canvas(self):  # pragma: no cover
        self.fig.canvas.draw()
    
    def plot_map_data(self, map_data, norm, cmap, extent):  # pragma: no cover
        # First clear the currently selected cursor
        self.map_clicked.emit(None, None)
        # Now plot the new data
        log.debug("Plotting new map data")
        self._latest_cmap = cmap
        self.map_ax.clear()
        if map_data.ndim == 2:
            # Scalar map showing some sort of metrich
            plots.plot_txm_map(map_data, ax=self.map_ax, norm=norm,
                               cmap=cmap, extent=extent)
        elif map_data.ndim == 3:
            # RGB map showing different components
            plots.plot_composite_map(map_data, ax=self.map_ax, extent=extent)
    
    def plot_histogram_data(self, map_data, norm, cmap, extent):  # pragma: no cover
        log.debug("Plotting new map histogram")
        self.hist_ax.clear()
        plots.plot_txm_histogram(map_data, ax=self.hist_ax, norm=norm,
                                 cmap=cmap, add_cbar=False)
        mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
        self.cbar.on_mappable_changed(mappable)
    
    def update_spectrum_ui(self, spectrum, fitted_spectrum, norm, cmap,
                      edge_range):  # pragma: no cover
        # Disable the edge_mask option if no edge range
        if edge_range is None:
            self.ui.chkEdgeMask.setEnabled(False)
        else:
            self.ui.chkEdgeMask.setEnabled(True)
    
    def plot_spectrum(self, spectrum, fitted_spectrum, norm, cmap,
                      edge_range):  # pragma: no cover
        log.debug("Plotting new map spectrum")
        # Clear the old axes
        self.spectrum_ax.clear()
        self.edge_ax.clear()
        # Plot the new spectra
        plots.plot_spectrum(
            spectrum=spectrum.values,
            energies=spectrum.index,
            norm=norm,
            color='x',
            cmap=cmap,
            ax=self.spectrum_ax,
        )
        plots.plot_spectrum(
            spectrum=spectrum.values,
            energies=spectrum.index,
            norm=norm,
            color='x',
            cmap=cmap,
            ax=self.edge_ax,
        )
        # Plot the fitted spectrum for comparison
        if fitted_spectrum is not None:
            self.spectrum_ax.plot(
                fitted_spectrum.index,
                fitted_spectrum.values,
                marker='None', linestyle=":")
            self.edge_ax.plot(
                fitted_spectrum.index,
                fitted_spectrum.values,
                marker='None', linestyle=":")
        # Set axes limits on the edge axes
        self.edge_ax.set_xlim(norm.vmin, norm.vmax)
        self.redraw_canvas()
    
    def connect_presenter(self, presenter):  # pragma: no cover
        """Connect to signals for changed presenter state."""
        presenter.map_data_changed.connect(self.show)
        presenter.map_data_changed.connect(self.plot_map_data)
        presenter.map_data_changed.connect(self.plot_histogram_data)
        presenter.map_data_changed.connect(self.redraw_canvas)
        presenter.map_data_cleared.connect(self.hide)
        presenter.map_spectrum_changed.connect(self.plot_spectrum)
        presenter.map_spectrum_changed.connect(self.update_spectrum_ui)
        presenter.cmap_list_changed.connect(self.set_cmap_list)
        presenter.component_list_changed.connect(self.set_component_list)
        presenter.map_limits_changed.connect(self.set_map_limits)
        presenter.map_cursor_changed.connect(self.update_cursor_labels)
        presenter.map_pixel_changed.connect(self.update_crosshair_labels)
        presenter.map_pixel_changed.connect(self.redraw_crosshairs)
        # Connect to signal for when we can draw the UI elements
        presenter.app_ready.connect(self.setup_ui)
        # `map_view` signals received when data changes
    
    def create_canvas(self):  # pragma: no cover
        self.fig = Figure()
        canvas = FigureCanvas(self.fig)
        self.ui.mainLayout.addWidget(canvas)
        canvas.setFocusPolicy(QtCore.Qt.StrongFocus)
        # Add subplots
        gridspec = GridSpec(2, 4)
        self.map_ax = self.fig.add_subplot(
            gridspec.new_subplotspec((0, 0), rowspan=2, colspan=2)
        )
        self.spectrum_ax = self.fig.add_subplot(
            gridspec.new_subplotspec((0, 2), rowspan=1, colspan=2)
        )
        self.hist_ax = self.fig.add_subplot(
            gridspec.new_subplotspec((1, 2), rowspan=1, colspan=1)
        )
        self.edge_ax = self.fig.add_subplot(
            gridspec.new_subplotspec((1, 3), rowspan=1, colspan=1)
        )
        # Create the colorbar on the histogram axes
        self.cbar = plots.draw_histogram_colorbar(ax=self.hist_ax,
                                                  cmap="viridis",
                                                  norm=Normalize(0, 1))
        self.cbar.ax.set_xlabel("Map Value")
        # Adjust the margins
        self.fig.tight_layout(pad=0)
        self.fig.canvas.draw_idle()
    
    def set_cmap_list(self, new_list):
        self.ui.cmbCmap.clear()
        self.ui.cmbCmap.addItems(new_list)
        # Check for, and disable, jet colormaps
        jet_idxs = [i for i, x in enumerate(new_list) if x == "jet"]
        for idx in jet_idxs:
            item = self.ui.cmbCmap.model().item(idx)
            item.setEnabled(False)
            item.setToolTip("Sorry, Brian...")
    
    def set_component_list(self, new_list):
        self.ui.cmbComponent.clear()
        self.ui.cmbComponent.addItems(new_list)
    
    def set_map_limits(self, vmin, vmax, step, decimals):
        log.debug('map view received map limits vmin=%f, vmax=%f, step=%f, decimals=%d',
                  vmin, vmax, step, decimals)
        # Temporarily disconnect signals
        oldStateVMin = self.ui.spnVMin.blockSignals(True)
        oldStateVMax = self.ui.spnVMax.blockSignals(True)
        # Update UI values
        # self.ui.spnVMin.setMinimum(vmin - 2 * step)
        self.ui.spnVMin.setMaximum(vmax - 10**-decimals)
        # self.ui.spnVMax.setMaximum(vmax + 2 * step)
        self.ui.spnVMax.setMinimum(vmin + 10**-decimals)
        self.ui.spnVMin.setDecimals(decimals)
        self.ui.spnVMax.setDecimals(decimals)
        self.ui.spnVMin.setSingleStep(step)
        self.ui.spnVMax.setSingleStep(step)
        self.ui.spnVMin.setValue(vmin)
        self.ui.spnVMax.setValue(vmax)
        # Reconnect any disconnected signals
        self.ui.spnVMin.blockSignals(oldStateVMin)
        self.ui.spnVMax.blockSignals(oldStateVMax)



# -*- coding: utf-8 -*-
#
# Copyright © 2017 Mark Wolf
#
# This file is part of Xanespy.
#
# Xanespy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Xanespy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Xanespy. If not, see <http://www.gnu.org/licenses/>.

"""A collection of callables that can be used for fitting spectra
against."""


from collections import namedtuple
import warnings

import numpy as np
from scipy.optimize import leastsq

from xanes_math import k_edge_jump, iter_indices, foreach


def prepare_p0(p0, frame_shape, num_timesteps=1):
    """Create an initial parameter guess for fitting.
    
    Takes a starting guess (p0) and returns a numpy array with this
    inital guess that matches the frameset.
    
    For example, if a frameset has 12 timesteps and (1024, 2048) frames,
    then a 5-tuple input for ``p0`` will result in a return value with
    shape (12, 5, 1024, 2048)
    
    """
    # Prepare an empty array for the results
    out_shape = (num_timesteps, *frame_shape, len(p0))
    out = np.empty(shape=out_shape)
    # Now populate the fields and put the param axis in the energy spot
    out[:] = p0
    out = np.moveaxis(out, -1, 1)
    return out


def fit_spectra(observations, func, p0, nonnegative=False, quiet=False):
    """Fit a function to a series observations.
    
    The shapes of ``observations`` and ``p0`` parameters must match in
    the first dimension, and the callable ``func`` should take a
    series of parameters (the exact number is determined by the last
    dimension of ``p0``) and return a set of observations (the length
    of which is determined by the last dimension of ``observations``).
    
    Parameters
    ----------
    observations : np.ndarray
      A 1- or 2-dimensional array of observations against which to fit
      the function ``func``.
    func : callable, str
      The function that will be used for fitting. It should match
      ``func(p0, p1, ...)`` where p0, p1, etc are the fitting
      parameters. Some useful functions can be found in the
      ``xanespy.fitting`` module.
    p0 : np.ndarray
      Initial guess for parameters, with similar dimensions to a
      frameset. Example, fitting 3 sources (plus offset) for a (1,
      40, 256, 256) 40-energy frameset requires p0 to be (1, 4,
      256, 256).
    nonnegative : bool, optional
      If true (default), negative parameters will be avoided. This
      can also be a tuple to allow for fine-grained control. Eg:
      (True, False) will only punish negative values in the first
      of the two parameters.
    quiet : bool, optional
      Whether to suppress the progress bar, etc.
    
    Returns
    -------
    params : numpy.ndarray
      The fit parameters (as frames) for each source.
    residuals : numpy.ndarray
      Residual error after fitting, as maps.

    """
    # Massage the datas
    observations = np.array(observations)
    p0 = np.array(p0)
    if observations.ndim == 1:
        residuals = np.zeros(shape=(1,))
    else:
        residuals = np.zeros(shape=(observations.shape[0],))
    params = np.empty_like(p0)
    # Prepare error function
    def errfun(guess, obs):
        if np.any(np.logical_and(guess < 0, nonnegative)):
            # Punish negative values
            diff = np.empty_like(obs)
            diff[:] = 1e6
        else:
            # Calculate error for this guess
            predicted = func(*guess)
            # Compare predicted with observed values
            diff = np.abs(obs - predicted)
        assert not np.any(np.isnan(diff))
        return diff
    # Execute fitting for each spectrum
    indices = iter_indices(observations, desc="Fitting spectra",
                           leftover_dims=1, quiet=quiet)
    def fit_sources(idx):
        spectrum = observations[idx]
        print(spectrum)
        # Don't bother fitting if there's NaN values
        if np.any(np.isnan(spectrum)):
            params[idx] = np.nan
            residuals[idx] = np.nan
            return
        # Valid data, so fit the spectrum
        guess = tuple(p0[idx])
        results = leastsq(func=errfun, x0=guess, args=(spectrum,), full_output=True)
        x, cov_x, infodict, mesg, status = results
        # Status 4 is often a sign of mismatched datatypes.
        if status == 4:
            msg = "Precision errors encountered during fitting. Check dtypes."
            warnings.warn(msg, RuntimeWarning)
        params[idx] = x
        # Calculate residual errors
        res_ = (spectrum - func(*x))
        res_ = np.sqrt(np.mean(np.power(res_, 2)))
        residuals[idx] = res_
    foreach(fit_sources, indices, threads=1)
    return (params, residuals)


class Curve():
    """Base class for a callabled Curve."""
    name = "curve"
    param_names = ()


class LinearCombination(Curve):
    """Combines other curves into one callable.
    
    The constructor accepts the keyword argument ``sources``, which
    should be a list of numpy arrays. The resulting object can then be
    called with parameters for the weight of each function plus an
    offset. For example, with two sources, the object is called as
    
    .. code:: python
    
        # Prepare the separate sources
        x = np.linspace(0, 2*np.pi, num=361)
        sources = [np.sin(x), np.sin(2*x)]
        
        # Produce a combo with 0.5*sin(x) + 0.25*sin(2x) + 2
        lc = LinearCombination(sources=sources)
        out = lc(0.5, 0.25, 2)
    
    The final output will have the same shape as the sources, which
    should all be the same shape as each other.
    
    """
    name = "linear_combination"
    
    def __init__(self, sources):
        self.sources = sources
    
    def __call__(self, *params):
        # Prepare data and parameters
        out = 0
        p_sources = params[0:-1]
        # Add each source weighted by input parameters
        for coeff, source in zip(p_sources, self.sources):
            out += coeff * source
        # Add global offset
        out += params[-1]
        return out
    
    @property
    def param_names(self):
        names = ['weight_%d' % idx for idx in range(len(self.sources))]
        names.append('offset')
        names = tuple(names)
        return names


class Gaussian(Curve):
    """A Gaussian curve.
    
    Mathematically:
    
    .. math::
        a e^{\frac{-(x-b)**2}{2c^2)}
    
    Parameters
    ----------
    x : np.ndarray
      Array of x-values to input into the Gaussian function.
    
    """
    name = "gaussian"
    param_names = ('height', 'center', 'width')
    
    def __init__(self, x):
        self.x = x
    
    def __call__(self, height, center, width):
        x = self.x
        a, b, c = (height, center, width)
        return a * np.exp(-(x-b)**2 / 2 / c**2)


class L3Curve(Curve):

    """An L_3 absorption edge.
    
    This function is a combination of two Gaussian peaks and a step
    function. The first 3 parameters give the height, position and
    width of one peak, and parameters 3:6 give the same for a second
    peak. Parameters 6:9 are height, position and width of an arctan
    step function. Parameter 9 is a global offset.
    
    Parameters
    ----------
    peaks : int, optional
      How many peaks to fit across the edge.
    
    """
    name = "L3-gaussian"
    
    def __init__(self, energies, num_peaks=2):
        self.num_peaks = num_peaks
        self.params = namedtuple('L3Params', self.param_names)
        self.energies = energies
        self.dtype = energies.dtype
    
    def __call__(self, *params):
        p = self.params(*params)
        Es = self.energies
        # Add two gaussian fields
        out = np.zeros_like(Es)
        gaussian = Gaussian(x=Es)
        for idx in range(self.num_peaks):
            i = 3*idx
            p_i = p[i:i+3]
            out += gaussian(*p_i)
        # Add arctan step function
        out += p.sig_height * (np.arctan((Es-p.sig_center)*p.sig_sigma) / np.pi + 0.5)
        # Add vertical offset
        out += p.offset
        return out
    
    @property
    def param_names(self):
        pnames = []
        # Add Gaussian parameters
        for idx in range(self.num_peaks):
            pnames.append('height_%d' % idx)
            pnames.append('center_%d' % idx)
            pnames.append('sigma_%d' % idx)
        # Add sigmoid parameters
        pnames.append('sig_height')
        pnames.append('sig_center')
        pnames.append('sig_sigma')
        # Add global y-offset parameter
        pnames.append('offset')
        return tuple(pnames)


class KCurve(Curve):
    param_names = (
        'scale', 'voffset', 'E0',  # Global parameters
        'sigw',  # Sharpness of the edge sigmoid
        'bg_slope', # Linear reduction in background optical_depth
        'ga', 'gb', 'gc',  # Gaussian height, center and width
    )
    
    def __init__(self, energies):
        self.energies = energies
    
    def guess_params(self, intensities, edge):
        """Guess initial starting parameters for a k-edge curve. This will
        give a rough estimate, appropriate for giving to the fit_kedge
        function as the starting parameters, p0.
        
        Arguments
        ---------
        intensities : np.ndarray
          An array containing optical_depth data that represents a
          K-edge spectrum. Only 1-dimensional data are currently
          accepted.
        edge : xanespy.edges.KEdge
          An X-ray Edge object, will be used for estimating the actual
          edge energy itself.
        
        Returns
        -------
        p0 : tuple
          A named tuple with the estimated parameters (see KEdgeParams
          for definition)
        
        """
        Is = intensities
        assert Is.shape == self.energies.shape
        # Guess the overall scale and offset parameters
        scale = k_edge_jump(frames=Is, energies=self.energies, edge=edge)
        voffset = np.min(Is)
        # Estimate the edge position
        E0 = edge.E_0
        # Estimate the whiteline Gaussian parameters
        ga = 5 * (np.max(Is) - scale - voffset)
        gb = self.energies[np.argmax(Is)] - E0
        gc = 2  # Arbitrary choice, should improve this in the future
        # Construct the parameters tuple
        KParams = namedtuple('KParams', self.param_names)
        p0 = KParams(scale=scale, voffset=voffset, E0=E0,
                             sigw=0.5, bg_slope=0,
                             ga=ga, gb=gb, gc=gc)
        return p0
    
    def __call__(self, *params):
        # Named tuple to help keep track of parameters
        Params = namedtuple('Params', self.param_names)
        p = Params(*params)
        x = self.energies
        # Adjust the x's to be relative to E_0
        x = x - p.E0
        # Sigmoid
        sig = np.arctan(x*p.sigw) / np.pi + 1/2
        # Gaussian
        gaus = p.ga*np.exp(-(x-p.gb)**2/2/p.gc**2)
        # Background
        bg = x * p.bg_slope
        curve = sig + gaus + bg
        curve = p.scale * curve + p.voffset
        return curve



# -*- coding: utf-8 -*-
#
# Copyright © 2016 Mark Wolf
#
# This file is part of Xanespy.
#
# Xanespy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Xanespy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Xanespy.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for setting up and displaying plots using matplotlib."""

import numpy as np
from matplotlib import pyplot, cm, rcParams
from matplotlib.colors import Normalize
from matplotlib.ticker import ScalarFormatter

from utilities import pixel_to_xy, Extent, Pixel


def latexify():
    """Set some custom options for saving matplotlib graphics in PGF
    format."""
    # Set default LaTeX PGF style
    pgf_with_latex = {                      # setup matplotlib to use latex for output# {{{
        "pgf.texsystem": "xelatex",        # change this if using xetex or lautex
        # "font.family": "serif",
        "font.serif": [],                   # blank entries should cause plots 
        "font.sans-serif": [],              # to inherit fonts from the document
        "font.monospace": [],
        "pgf.preamble": [
            r"\usepackage[utf8x]{inputenc}",    # use utf8 fonts 
            r"\usepackage[T1]{fontenc}",        # plots will be generated
            r"\usepackage{fontspec}",
            r"\usepackage[detect-all,locale=DE,per-mode=reciprocal]{siunitx}",
            r"\usepackage[version=4]{mhchem}",
        ]
    }
    rcParams.update(pgf_with_latex)
    # Default to tight bounding box
    # rcParams['savefig.bbox'] = "tight"


def remove_extra_spines(ax):
    """Removes the right and top borders from the axes."""
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.yaxis.set_ticks_position('left')
    ax.xaxis.set_ticks_position('bottom')
    return ax


def set_axes_color(ax, color):  # pragma: no cover
    """Set the axes, tick marks, etc of `ax` to mpl color `color`. Also,
    "doegreen" has special significance as the color associated with the
    US department of energy."""
    if color == "doegreen":
        color = (33/255, 99/255, 50/255)
    # Set the spine colors
    ax.spines['bottom'].set_color(color)
    ax.spines['top'].set_color(color)
    ax.spines['right'].set_color(color)
    ax.spines['left'].set_color(color)
    # Change ticks colors
    ax.tick_params(axis='x', colors=color)
    ax.tick_params(axis='y', colors=color)
    # Change label color
    ax.yaxis.label.set_color(color)
    ax.xaxis.label.set_color(color)


def set_outside_ticks(ax):  # pragma: no cover
    """Convert all the axes so that the ticks are on the outside and don't
    obscure data."""
    ax.get_yaxis().set_tick_params(which='both', direction='out')
    ax.get_xaxis().set_tick_params(which='both', direction='out')
    return ax


def new_axes(height=5, width=None):
    """Create a new set of matplotlib axes for plotting. Height in inches."""
    # Adjust width to accomodate colorbar
    if width is None:
        width = height / 0.8
    fig = pyplot.figure(figsize=(width, height))
    # Set background to be transparent
    fig.patch.set_alpha(0)
    # Create axes
    ax = pyplot.gca()
    # Remove borders
    remove_extra_spines(ax)
    return ax


def new_image_axes(height=5, width=5):  # pragma: no cover
    """Square axes with ticks on the outside."""
    ax = new_axes(height, width)
    return set_outside_ticks(ax)


def big_axes():  # pragma: no cover
    """Return a new Axes object, but larger than the default."""
    return new_axes(height=9, width=16)


def dual_axes(fig=None, longdim=13.8, shortdim=6.9, orientation='horizontal'):  # pragma: no cover
    """Two new axes for mapping, side-by-side.

    Parameters
    ----------
    longdim : float
      Size in inches for the long dimension. If orientation is
      "vertical", this will be the height.
    shortdim : float
      Size in inches for the short dimension. If orientation is
      "vertical", this will be the width.
    """
    if fig is None:
        fig = pyplot.figure()
    if orientation == 'vertical':
        fig, (ax1, ax2) = fig.subplots(2, 1)
        fig.set_figwidth(shortdim)
        fig.set_figheight(longdim)
    else:
        fig, (ax1, ax2) = pyplot.subplots(1, 2)
        fig.set_figwidth(longdim)
        fig.set_figheight(shortdim)
    # Remove redundant borders
    remove_extra_spines(ax1)
    remove_extra_spines(ax2)
    # Set background to be transparent
    fig.patch.set_alpha(0)
    return (ax1, ax2)


def draw_histogram_colorbar(ax, *args, **kwargs):  # pragma: no cover
    """Similar to `draw_colorbar()` with some special formatting options
    to put it along the X-axis of the axes."""
    cbar = draw_colorbar(ax=ax, pad=0, orientation="horizontal", energies=None, *args, **kwargs)
    ax.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='off',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        labelbottom='off',
        labeltop='off')
    ax.spines['bottom'].set_visible(False)
    cbar.ax.set_xlabel(ax.get_xlabel())
    ax.xaxis.set_visible(False)
    cbar.ax.set_title("")
    cbar.outline.set_visible(False)
    gray = (0.1, 0.1, 0.1)
    cbar.ax.axhline(cbar.ax.get_ylim()[1], linewidth=2, linestyle=":", color=gray)
    cbar.ax.tick_params(
        axis='x',
        which='both',
        bottom='on',
        top='on',
        labelbottom="on",
    )
    return cbar


def draw_colorbar(ax, cmap, norm, energies=None, orientation="vertical",
                  *args, **kwargs):  # pragma: no cover
    """Draw a colorbar on the side of a mapping axes to show the range of
    colors used. Returns the newly created colorbar object.

    Parameters
    ---------
    ax
      Matplotlib axes object against which to plot.
    cmap : str
      String or mpl Colormap instance indicating which colormap to
      use.
    norm : matplotlib.Normalize
      Describes the range of values to use.
    energies : iterable, optional
      Values to put as the tick marks on the colorbar. If not given, 3
      points across ``norm`` will be used.
    orientation : str, optional
      "horizontal" or "vertical" (default)
    
    """
    mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
    mappable.set_array(np.arange(0, 3))
    # Prepare default energies if necessary
    if energies is None:
        energies = np.linspace(norm.vmin, norm.vmax, num=3)
    # Add the colorbar to the axes
    cbar = pyplot.colorbar(mappable,
                           ax=ax,
                           ticks=energies,
                           spacing="proportional",
                           orientation=orientation,
                           *args, **kwargs)
    # Annotate the colorbar
    cbar.ax.set_title('eV')
    # Make sure the ticks don't use scientific notation
    cbar.formatter.set_useOffset(False)
    cbar.update_ticks()
    return cbar


def plot_pixel_spectra(pixels, extent, spectra, energies, map_ax,
                       spectra_ax=None, step_size=0):  # pragma: no cover
    """Highlight certain pixels in an already-plotted map and plot their
    spectra. The map should already have been plotted.

    Arguments
    ---------

    - pixels : List of 2-tuples with (row, column) positions of pixels
      to highlight and plot.

    - extent : Matplotlib extent specification for converting (y, x)
      positions to (row, column) positions.

    - spectra : Numpy array with spectra that will be plotted. Entries
      in `pixels` will use this array to get spectra. Shape is assumed
      to be (row, column, energy).

    - energies : Numpy array with energies to use as the x-axis for
      plotting spectra.

    - map_ax : The matplotlib Axes that will be used for highlighting
      pixels.

    - spectra_ax : A matplotlib Axes on which to plot the spectra. If
      None (default) a new axes will be created.

    - step_size : An offset to add to each spectrum so they are not
      directly on top of each other.
    """
    # Prepare some metadata
    map_shape = spectra.shape[0:2]
    extent = Extent(*extent)
    # Prepare axes if necessary (and remove spines for arbitrary units)
    if spectra_ax is None:
        spectra_ax = new_axes()
    remove_extra_spines(spectra_ax)
    # Remove ticks on the y axes
    tick_params = {
        'axis': 'y',
        'which': 'both',
        'left': 'off',
        'right': 'off',
        'labelright': 'off',
        'labelleft': 'off',
    }
    # Prepare a secondary y axes if necessary for complex data
    if np.iscomplexobj(spectra):
        spectra_ax2 = spectra_ax.twinx()
        remove_extra_spines(spectra_ax2)
        spectra_ax2.tick_params(**tick_params)
    spectra_ax.tick_params(**tick_params)
    # Do the plotting and annotating for each pixel
    min_val = 0
    max_val = 0
    for idx, px in enumerate(pixels):
        # Put an annotation on the map
        px = Pixel(*px)
        xy = pixel_to_xy(px, extent, shape=map_shape)
        map_ax.text(xy.x, xy.y, str(idx), color="white",
                    horizontalalignment="center", verticalalignment="center",)
        # Normalize the spectrum
        spectrum = spectra[px]
        min_ = np.min(np.abs(spectrum))
        max_ = np.max(np.abs(spectrum))
        spectrum = (spectrum - min_) / (max_ - min_)
        # Plot the spectrum on the spectrum axes
        spectrum += step_size * idx
        plot_spectrum(spectrum, energies, ax=spectra_ax,
                      ax2=spectra_ax2)
        # Put labels on the axes to indicate which specturm is which
        spectra_ax.text(
            x=np.max(energies) + 2,
            y=np.mean((np.real(spectrum), np.imag(spectrum))),
            s=idx,
        )
        # Save values for proper scaling of axes
        min_val = complex(
            min(np.real(min_val), np.real(spectrum).min()),
            min(np.imag(min_val), np.imag(spectrum).min())
        )
        max_val = complex(
            max(np.real(max_val), np.real(spectrum).max()),
            max(np.imag(max_val), np.imag(spectrum).max())
        )
    # Set axes limits based on min and max values
    min_val -= min_val * 0.1
    max_val += max_val * 0.1
    spectra_ax.set_ylim(bottom=np.real(min_val), top=np.real(max_val))
    if spectra_ax2 is not None:
        spectra_ax2.set_ylim(bottom=np.imag(min_val), top=np.imag(max_val))
    spectra_ax.set_xlim(right=np.max(energies)+4)

def plot_kedge_fit(energies, params):
    """Plot the fit based on the given k-edge params. The params will be
    given to ``xanespy.xanes_math.KEdgeParams`` for conversion.

    energies : np.ndarray
      A 1D array with the x values to plot.

    params : array-like
      Fitted K-edge parameters.
    """
    p = KEdgeParams(*params)
    # Plot the sigmoid by making everything else zero
    sig_p = dict(zip(kedge_params, p))
    sig_p['ga'] = 0
    sig_p = KEdgeParams(**sig_p)
    sig = predict_edge(energies, *sig_p)
    pyplot.plot(energies, sig, linestyle=":")
    # Plot the whole predicted line
    fit = predict_edge(energies, *p)
    pyplot.plot(energies, fit, linestyle=":")

def plot_spectrum(spectrum, energies, norm=Normalize(),
                  show_fit=False, ax=None, ax2=None,
                  linestyle=':', color="blue", cmap="plasma",
                  polar_coords=False,
                  *args, **kwargs):  # pragma: no cover
    """Plot an energy spectrum on an axes.
    
    Applies some color formatting if `edge` is a valid XANES Edge
    object.
    
    Arguments
    ---------
    spectrum : np.ndarray
      Array of intensity values.
    energies : np.ndarray
      Array of energy values.
    norm : optional
      Matplotlib Normalize() object that shows the map
      range. This will be used to annotate the plot if it is give.
    show_fit : bool, optional
      Whether to plot lines showing the best fit.
    ax : mpl.Axes, optional
      Matplotlib Axes on which to plot. If not given, a new axes
      will be generated.
    ax2 : mpl.Axes, optional
      A second y-axes for plotting the imaginary component if
      the data are complex.
    linestyle : optional
      Passed on to matplotlib.
    cmap : str, optional
      Colormap, passed on to matplotlib
    color : optional
      Specifies the color for the circles plotted. Either "x"
      or "y" will decide based on the numerical value, `norm` and
      `cmap` arguments. Anything else will be passed as a color spec
      to the matplotlib commands.
    polar_coords : bool, optional
      If truthy, the spectrum will be plotted as modulus-phase instead
      of real-imag. For purely real data this is equivalent to taking
      the absolute value.
    
    """
    if ax is None:
        ax = new_axes()
    if norm is not None:
        norm.autoscale_None(np.real(spectrum))
    # Retrieve `values` in case it's a pandas series
    spectrum = getattr(spectrum, 'values', spectrum)
    # Color code the markers by energy
    is_complex = np.iscomplexobj(spectrum)
    if is_complex:
        colors = None
    elif color == "x":
        colors = cm.get_cmap(cmap)(norm(energies))
    elif color == "y":
        colors = cm.get_cmap(cmap)(norm(spectrum))
    else:
        colors = [color] * len(energies)
    # Remove secondary axes and re-add them for complex values
    if is_complex:
        # Check if modulus-phase or real-imag components
        if polar_coords:
            comp0, comp1 = (np.abs, np.angle)
            label0, label1 = ('Modulus', 'Phase')
        else:
            comp0, comp1 = (np.real, np.imag)
            label0, label1 = ('Real', 'Imag')
        # Create a secondary axis
        if ax2 is None:
            ax2 = ax.twinx()
        # Convert complex values to two lines
        ys = spectrum
        artist = ax.plot(energies, comp0(ys), linestyle=linestyle, color="C0")
        artist.extend(ax2.plot(energies, comp1(ys),
                               linestyle=linestyle, color="C1", *args, **kwargs))
    else:
        # Just plot the real numbers
        ys = np.real(spectrum)
        artist = ax.plot(energies, ys, linestyle=linestyle, *args, **kwargs)
    # save limits, since they get messed up by scatter plot
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    # Draw scatter plot of data points
    markersize = kwargs.get('markersize', rcParams['lines.markersize'])
    if is_complex:
        ax.scatter(energies, comp0(ys), c="C0", s=25, alpha=0.5)
        ax.set_ylabel(label0, color="C0")
        for t1 in ax.get_yticklabels():
            t1.set_color("C0")
        ax2.scatter(energies, comp1(ys), c="C1", s=markersize ** 2, alpha=0.5)
        ax2.set_ylabel(label1, color="C1")
        for t1 in ax2.get_yticklabels():
            t1.set_color("C1")
    else:
        ax.scatter(energies, ys, c=colors, s=markersize ** 2)
        ax.set_ylabel('Optical Depth')
    ax.set_xlim(*xlim)
    ax.set_ylim(*ylim)
    ax.set_xlabel('Energy /eV')
    # Plot the edges of the map range
    if norm is not None:
        vlineargs = dict(linestyle='--', alpha=0.7,
                         color="lightgrey", zorder=0,)
        ax.axvline(norm.vmin, **vlineargs)
        ax.axvline(norm.vmax, **vlineargs)
    return artist


def plot_composite_map(data, ax=None, origin="upper", *args, **kwargs):  # pragma: no cover
    """Plot an RGB composite map on the given axes."""
    while data.shape[-1] < 3:
        data = np.concatenate((data, np.zeros_like(data)),
                              axis=-1)
    data = data[..., 0:3]
    if ax is None:
        ax = new_image_axes()
    artist = ax.imshow(data, origin=origin, *args, **kwargs)
    return artist


def plot_txm_map(data, edge=None, norm=None, ax=None, cmap='plasma',
                 origin="upper", vmin=None, vmax=None,
                 *args, **kwargs):  # pragma: no cover
    # Get a default normalizer
    if norm is None:
        norm = Normalize(vmin, vmax)
    norm.autoscale_None(data[~np.isnan(data)])
    # Create axes if necessary
    if ax is None:
        ax = new_image_axes()
        if edge is None:
            # No specific edge is given, so use up to 10 evenly space energies
            num_Es = round(min(abs(norm.vmax - norm.vmin + 1), 10))
            energies = np.linspace(norm.vmin, norm.vmax, num=num_Es)
        else:
            energies = edge.energies_in_range(norm_range=(norm.vmin, norm.vmax))
        draw_colorbar(ax=ax, cmap=cmap, norm=norm, energies=energies)
    # Do the plotting
    artist = ax.imshow(data,
                       cmap=cmap,
                       origin=origin,
                       norm=norm,
                       *args, **kwargs)
    # Add annotations and formatting stuff
    return artist


def plot_txm_histogram(data, ax=None, norm=None, bins=None,
                       cmap='plasma', add_cbar=True, *args, **kwargs):  # pragma: no cover
    """Take an array of data values and show a histogram with some
    color-coding related to normalization value.

    Returns: The matplotlib axes object used for plotting.

    Parameters
    ----------
    data : np.ndarray
      An array of values to plot on the histogram.
    ax : optional
      Matplotlib Axes instance to receive the plot. If None, a new axes
      will created.
    norm : optional
      Matplotlib Normalize instance with the colormap range.
    bins : optional
      Bins to pass to the matplotlib hist() routine. If None
      (default), we will choose based on dtype of the data: integers
      will yield 1-wide bins, anything else will give 256 bins.
    cmap : str, optional
      Matplotlib colormap for coloring the bars.
    add_cbar : bool, optional
      Boolean to decide whether to add a colorbar along the
      bottom axis or not.
    *args
      Positional arguments passed to matplotlib's `hist` call.
    *kwargs
      Keyword arguments passed to matplotlib's `hist` call.
    
    """
    if ax is None:
        ax = new_axes()
    # Flatten the data so it can be nicely plotted
    if hasattr(data, 'mask'):
        data = data[~data.mask]
    else:
        data = data.flatten()
    # Remove any Not-a-number values
    data = data[~np.isnan(data)]
    # Set normalizer
    if norm is None:
        norm = Normalize()
        norm.autoscale_None(data)
    # Clip the data so that it includes the end-members
    clip_data = np.clip(data, norm.vmin, norm.vmax)
    unique_vals = np.unique(np.array(clip_data))
    # Determine a reasonable binning parameter based on data type
    if bins is not None:
        pass
    elif np.issubdtype(data.dtype, np.integer):
        # integers, so make the bins 1-wide
        bins = np.arange(norm.vmin, norm.vmax, step=1)
    elif len(unique_vals) < 256:
        # Not many values to choose from, so use the values as bins
        bins = np.sort(unique_vals)
    else:
        bins = 256
    # Plot the histogram
    n, bins, patches = ax.hist(clip_data, bins=bins, *args, **kwargs)
    # Set colors on histogram
    for patch in patches:
        x_position = patch.get_x()
        real_cmap = cm.get_cmap(cmap)
        color = real_cmap(norm(x_position))
        patch.set_color(color)
    # Set axes decorations
    ax.set_ylabel("Pixels")
    ax.set_xlim(norm.vmin, norm.vmax)
    # ax.xaxis.set_ticks(energies)
    ax.xaxis.get_major_formatter().set_useOffset(False)
    # Draw a colorbar along the bottom axes
    if add_cbar:
        cbar = draw_histogram_colorbar(ax=ax, cmap=cmap, norm=norm)
    else:
        cbar = None
    return (ax, cbar)


def plot_spectra(spectra, energies, ax=None):
    """Take an iterable of spectra and plot them one above the next.
    
    Parameters
    ----------
    spectra : iterable
      Each entry is an array with intensity values.
    energies : np.ndarray
      Energy values for the points in each entry of ``spectra``.
    ax : matplotlib.Axes, optional
      The Axes object to receive the plots. If omitted, a new Axes
      will be created.
    
    Returns
    -------
    artists : list
      A list of the matplotlib artists used to draw the spectra.
    
    """
    artists = []
    # Get a default axes if one was not provided
    if ax is None:
        ax = new_axes()
    # Do the plotting for each spectrum
    yoffset = 0
    for spectrum in spectra:
        artist = ax.plot(energies, spectrum + yoffset, marker='o', linestyle=':')
        artists.append(artist)
        # Calculate the appropriate step for the next round
        yoffset += np.median(spectrum) - 0.25 * (np.max(spectrum) - np.min(spectrum))
    return artists


def plot_spectra_as_map(spectra, energies, ax=None, extent=None, **kwargs):
    """Take an iterable of spectra and plot them as a heat map.
    
    This function takes energies so it can put ticks on the x-axis,
    however if they are not equally spaced, their distance on the map
    will not properly capture they're distance numerically.
    
    Parameters
    ----------
    spectra : iterable
      Each entry is an array with intensity values.
    energies : np.ndarray
      Energy values for the points in each entry of ``spectra``.
    ax : matplotlib.Axes, optional
      The Axes object to receive the plots. If omitted, a new Axes
      will be created.
    extent : Matplotlib extent.
    **kwargs : Get passed on to matplotlib imshow().
    
    Returns
    -------
    artists : list
      A list of the matplotlib artists used to draw the spectra.
    
    """
    # Get a default axes if one was not provided
    if ax is None:
        ax = new_axes()
    # Do the plotting for each spectrum
    artist = ax.imshow(spectra, origin='lower', extent=extent, aspect="auto", **kwargs)
    return [artist]


def plot_txm_intermediates(images):  # pragma: no cover
    """Accept a dictionary of images and plots them each on its own axes
    using matplotlib's `imshow`. This is a complement to routines that
    operate on a microscopy frame and optionally return all the
    intermediate calculated frames.
    
    """
    for key in images.keys():
        ax1, ax2 = dual_axes()
        ax1.imshow(images[key], cmap='gray')
        ax1.set_title(key)
        ax2.hist(images[key].flat, bins=100)



# -*- coding: utf-8 -*-
#
# Copyright © 2016 Mark Wolf
#
# This file is part of Xanespy.
#
# Xanespy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Xanespy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Xanespy. If not, see <http://www.gnu.org/licenses/>.

"""Module containing all the computationally demanding functions. This
allows for easy optimization of parallelizable algorithms and better
test isolation. Most functions will operate on large arrays of data.

"""

import warnings
import logging
from time import time
import multiprocessing as mp
from itertools import product
from collections import namedtuple

from scipy import ndimage, linalg, stats
from scipy.optimize import curve_fit, leastsq
import numpy as np
from skimage import transform, feature, filters, morphology, exposure, measure
from sklearn import decomposition
import matplotlib.pyplot as plt
from tqdm import tqdm

from utilities import prog, foreach, get_component
from exceptions import XanesMathError


log = logging.getLogger(__name__)


# Helpers for parallelizable code
def iter_indices(data, leftover_dims=1, desc=None, quiet=False, ):
    """Accept an array of frames, indices, etc. and generate slices for
    each frame.
    
    Assumes the last two dimensions of `data` are rows and
    columns. All other dimensions will be iterated over.
    
    Parameters
    ----------
    data : np.ndarray
      Data to iterate over.
    leftover_dims : int, optional
      Integer describing which dimensions should not be iterated
      over. Eg. if data is 3D array and leftover_dims == 1, only first
      two dimenions will be iterated.
    desc : str, optional
      String to put in the progress bar.
    quiet : bool, optional
      Whether to suppress the progress bar, etc.
    
    """
    fs_shape = np.asarray(data.shape[:-leftover_dims])
    ranges = [range(0, s) for s in fs_shape]
    indices = product(*ranges)
    if not quiet:
        indices = prog(indices, desc=desc, total=np.product(fs_shape))
    return indices


def downsample_array(arr, factor, method='mean', axis=None):
    """Reduced the shape of an image in powers of 2.
    
    Increases in the signal-to-noise can be achieved by sacrificing
    spatial resolution: a (1024, 1024) image can be converted to (512,
    512) by taking the mean of each (2, 2) block. The ``factor``
    parameter controls how agressive this conversion is. Extra pixels
    get dropped: eg (1025, 1024) -> (512, 512).
    
    +--------------+--------+--------+--------------+
    | arr          | factor | block  | out          |
    +--------------+--------+--------+--------------+
    | (1024, 1024) | 0      | N/A    | (1024, 1024) |
    | (1024, 1024) | 1      | (2, 2) | (512, 512)   |
    | (1024, 1024) | 2      | (4, 4) | (256, 256)   |
    | (1024, 1024) | 3      | (8, 8) | (128, 128)   |
    +--------------+--------+--------+--------------+
    
    Parameters
    ----------
    arr : np.ndarray
      Input array to be downsampled.
    factor : int
      Downsampling factor. A value of 0 returns the original array.
    method : str, optional
      How to convert a (eg 2x2) block to a single value. Valid choices
      are 'mean' (default), 'median', 'sum'.
    axis : tuple, optional
      Which axes to use for reduction. If omitted, all axes will be
      used.
    
    Raises
    ------
    ValueError
      Negative downsampling factors.
    ValueError
      ``method`` parameter is not one of the valid options.
    
    Returns
    -------
    out : np.ndarray
      The downsample array.

    """
    # Check that the method parameter is valid
    methods = {
        'mean': np.mean,
        'sum': np.sum,
        'median': np.median,
    }
    if method not in methods.keys():
        raise ValueError("Invalid method '{}'. Choices are {}"
                         "".format(method, list(methods.keys())))
    # Prepare a list of axes to operate on if not given
    if axis is None:
        axis = tuple(range(arr.ndim))
    # Check that the downsampling factor is sane.
    if factor < 0:
        raise ValueError("Downsampling factor must be non-negative: {}"
                         "".format(factor))
    elif factor == 0:
        # No resampling necessary
        out = arr
    else:
        # Calculate new shapes and block sizes
        blocksize = 2**factor
        def subshape(arr, bs, axis):
            for idx in range(arr.ndim):
                if idx in axis:
                    yield int(arr.shape[idx] / bs) * bs
                else:
                    yield arr.shape[idx]
        # subshape = tuple(int(s / blocksize) * blocksize for s in arr.shape)
        subshape = tuple(subshape(arr, blocksize, axis))
        # Remove any extra pixels that don't fit into a block
        slices = tuple(slice(0, s, 1) for s in subshape)
        arr = arr[slices]
        # Create extra axes for reducing
        def tmpshape(arr, bs, axis):
            for idx in range(arr.ndim):
                if idx in axis:
                    # Yes downsample
                    yield int(arr.shape[idx] / bs)
                    yield bs
                else:
                    # No downsample
                    yield arr.shape[idx]
                    yield 1
        tmpshape = tuple(tmpshape(arr, blocksize, axis))
        arr = arr.reshape(tmpshape)
        # Reduce the extra (odd) axes to produce a downsampled image
        axis = tuple(range(1, arr.ndim, 2))
        out = methods[method](arr, axis=axis)
    return out


def apply_mosaic_reference(intensity, reference):
    """Use a single reference frame to calculate the reference for a
    mosaic of intensity frames.
    
    Returns
    -------
    - out : ndarray
      Same shape as ``intensity`` (I) but with optical depth: ln(I/I0)
      where I0 is the reference frame.
    
    Parameters
    ----------
    - intensity : ndarray
      Intensity data image of the sample. It's shape must be a whole
      multiple of the shape of ``reference``.
    - reference : ndarray
      A reference image with no sample that will be applied.
    
    """
    reps = np.divide(intensity.shape, reference.shape).astype(int)
    mosaic_ref = np.tile(reference, reps)
    od = np.log(mosaic_ref / intensity)
    return od


def apply_internal_reference(intensities, out=None, quiet=False):
    """Apply a reference correction to complex data to convert intensities
    into refractive index. :math:`I_0` is determined by separating the pixels
    into background and foreground using Otsu's method.
    
    Arrays `intensities` and `out` must all have the same shape where
    the last two dimensions are image rows and column.
    """
    is_complex = np.iscomplexobj(intensities)
    j = complex(0, 1)
    if is_complex:
        intensities = intensities.astype(np.complex128)
        component = "modulus"
    else:
        intensities = intensities.astype(np.float64)
        component = "real"
    logstart = time()
    log.debug("Starting internal reference correction")
    if out is None:
        out = np.empty_like(intensities)
    def apply_ref(idx):
        # Calculate background intensity using thresholding
        log.debug("Applying reference for frame %s", idx)
        frame = intensities[idx]
        if is_complex:
            direct_img = get_component(frame, component)
        else:
            direct_img = frame
        threshold = filters.threshold_yen(direct_img)
        graymask = direct_img > threshold
        background = frame[graymask]
        # Calculate an I_0 based on median modulus and phase
        if is_complex:
            mod = np.median(np.abs(background))
            angle = np.median(np.angle(background))
            j = complex(0, 1)
            I_0 = mod * np.cos(angle) + j * mod * np.sin(angle)
        else:
            I_0 = np.median(background)
        log.debug("I0 for frame %s = %f", idx, I_0)
        # Calculate optical depth based on background
        od = np.log(I_0 / frame)
        # if is_complex:
        #     # Calculate relative phase shift if complex data is required
        #     phase = np.angle(frame)
        #     phase - np.median((phase * graymask)[graymask > 0])
        #     # The phase data has a gradient in the background, so remove it
        #     x, y = np.meshgrid(np.arange(phase.shape[-1]),
        #                        np.arange(phase.shape[-2]))
        #     A = np.column_stack([y.flatten(),
        #                          x.flatten(),
        #                          np.ones_like(x.flatten())])
        #     p, residuals, rank, s = linalg.lstsq(A, phase.flatten())
        #     bg = p[0] * y + p[1] * x + p[2]
        #     # Prepare the complex output data
        #     phase = phase - bg
        #     j = complex(0, 1)
        #     out[idx] = phase + j*optical_depth
        # else:
        #     # Real data (optical-depth) only
        #     out[idx] = optical_depth
        out[idx] = od
        log.debug("Applied internal reference for frame %s", idx)
    # Call the actual function for each frame
    iter_frames = iter_indices(intensities, leftover_dims=2,
                               desc="Apply reference", quiet=quiet)
    foreach(apply_ref, iter_frames)
    log.info("Internal reference applied in %f seconds",
             (time() - logstart))
    return out


def extract_signals_nmf(spectra, n_components, nmf_kwargs=None, mask=None):
    """Extract the signal components present in the given spectra using
    non-negative matrix factorization. Input data can be negative, but
    it will be shifted up, processed, then shifted down again.
    
    Arguments
    ---------
    spectra : numpy.ndarray
      A numpy array of observations where the last axis is energy.
    n_components : int
      How many components to extract from the data.
    nmf_kwargs : dict, optional
      Keyword arguments to be passed to the constructor of the
      estimator.
    
    Returns
    -------
    components, weights : numpy.ndarray
      Extracted components and weights for each pixel-component
    combination.
    
    """
    if nmf_kwargs is None:
        _nmf_kwargs = {}
    max_iter = _nmf_kwargs.pop('max_iter', 200)
    # Make sure all the values are non-negative
    _spectra = np.abs(spectra)
    # Perform NMF fitting
    nmf = decomposition.NMF(n_components=n_components,
                            max_iter=max_iter,
                            **_nmf_kwargs)
    weights = nmf.fit_transform(_spectra)
    # Extract results and calculate weights
    signals = nmf.components_
    # Log the results
    log.info("NMF of %d samples in %d of %d iterations.", len(spectra),
             nmf.n_iter_+1, max_iter)
    log.info("NMF reconstruction error = %f", nmf.reconstruction_err_)
    return signals, weights


def apply_references(intensities, references, out=None, quiet=False):
    """Apply a reference correction to convert intensity values to optical
    depth.
    
    The formula :math:`-ln\\frac{intensities}{references}` is used to
    calculate the new values. Arrays ``intensities``, ``references``
    and ``out`` must all have the same shape where the last two
    dimensions are image rows and columns.
    
    Parameters
    ----------
    intensities : np.ndarray
      Sample input signal data.
    references : np.ndarray
      Background input signal data. Must be the same shape as
      intensities.
    out : np.ndarray, optional
      Array to receive the results.
    quiet : bool, optional
      Whether to suppress progress bar, etc.
    """
    logstart = time()
    # Create an empty array to hold the results
    if out is None:
        out = np.zeros_like(intensities)
    assert intensities.shape == references.shape
    # Reshape to be an array of images
    image_shape = (intensities.shape[-2], intensities.shape[-1])
    Is = np.reshape(intensities, (-1, *image_shape))
    refs = np.reshape(references, (-1, *image_shape))
    out_shape = out.shape[:-2]
    # Function for parallelization
    def calc(idx):
        out_idx = np.unravel_index(idx, out_shape)
        out[out_idx] = -np.log(np.divide(Is[idx], refs[idx]))
    prog_iter = prog(range(0, Is.shape[0]), desc="Calc'ing OD",
                     disable=quiet)
    foreach(calc, prog_iter)
    return out


def l_edge_mask(frames: np.ndarray, energies: np.ndarray, edge,
                sensitivity: float=1, frame_dims=2, min_size=0):
    """Calculate a mask for what is likely active material at this
    edge. This is done by comparing each spectrum to the overall
    spectrum using the dot product. A normalization is first applied
    to mitigate differences in total intensity.
    
    Parameters
    ---------
    frames : np.ndarray
      Array with images at different energies.
    energies : np.ndarray
      X-ray energies corresponding to images in `frames`. Must have
      the same shape along the first dimenion as `frames`.
    edge : xanespy.edges.Edge
      An Edge object that contains a description of the elemental edge
      being studied.
    sensitivity : float, optional
      A multiplier for the otsu value to determine the actual
      threshold.
    frame_dims : int, optional
      The number of dimensions that each frame has. Eg, 2 means each
      frame is a two-dimensional image.
    min_size : float, optional
      Objects below this size (in pixels) will be
      removed. Passing zero (default) will result in no effect.
    
    Returns
    -------
    mask : np.ndarray
      A boolean mask with the same shape as the last two dimensions of
      `frames` where True pixels are likely to be background material.
    
    """
    # Take the mean over all timesteps
    ODs = frames
    if frames.ndim > 3:
        E_dim = frame_dims + 1
        ODs = np.mean(np.reshape(frames, (-1, *frames.shape[-E_dim:])), axis=0)
        Es = np.mean(np.reshape(energies, (-1, frames.shape[-E_dim])), axis=0)
    elif frames.ndim == 3:
        Es = energies
    # Convert optical depths into spectra -> (spectrum, energy) shape
    frame_shape = frames.shape[-frame_dims:]
    spectra = ODs.reshape(-1, np.prod(frame_shape))
    spectra = np.moveaxis(spectra, 0, 1)
    assert spectra.ndim == 2
    # Compute normalized global average spectrum
    spectrum = np.mean(spectra, axis=0)
    global_min = np.min(spectrum)
    global_max = np.max(spectrum)
    spectrum = (spectrum - global_min) / (global_max - global_min)
    # Compute some statistical values for normalization
    minima = np.min(spectra, axis=1)
    maxima = np.max(spectra, axis=1)
    # Compute the spectrum baseline
    pre_edge_mask = np.logical_and(
        np.min(edge.pre_edge) <= Es,
        Es <= np.max(edge.pre_edge),
    )
    post_edge_mask = np.logical_and(
        np.min(edge.post_edge) <= Es,
        Es <= np.max(edge.post_edge),
    )
    # Check the pre- and post-edge regions are in the data
    if not (np.any(pre_edge_mask) and np.any(post_edge_mask)):
        msg = "Cannot find pre and post edges for L-edge: {}"
        msg = msg.format(edge)
        warnings.warn(msg, RuntimeWarning)
    baseline_idx = np.logical_or(pre_edge_mask, post_edge_mask)
    baseline = np.mean(spectrum[baseline_idx])
    # Normalize all the pixel spectra based on the baselines and maxima
    pixels = np.moveaxis(spectra, 0, 1)
    pixels = (pixels - baseline) / (maxima - minima)
    # Compute the overlap between the global and pixel spectra
    overlap = np.dot(spectrum, pixels)
    # Threshold the pixel overlap to find foreground vs background
    threshold = filters.threshold_otsu(overlap) * sensitivity
    mask = overlap > threshold
    # Recreate the original image shape as a boolean array
    mask = np.reshape(mask, frame_shape)
    # Remove left-over background fuzz
    if min_size > 0:
        mask = morphology.opening(mask, selem=morphology.disk(min_size))
    mask = np.logical_not(mask)
    return mask


def k_edge_mask(frames: np.ndarray, energies: np.ndarray, edge,
                sensitivity: float=1, min_size=0):
    """Calculate a mask for what is likely active material at this
    edge. This is done by comparing the edge-jump to the standard
    deviation. Foreground material will be identified when the
    edge-jump accounts for most of the standard deviation.
    
    Arguments
    ---------
    frames : numpy.ndarray
      Array with images at different energies.
    energies : numpy.ndarray
      X-ray energies corresponding to images in ``frames``. Must have
      the same shape along the first dimenion as ``frames``.
    edge : KEdge
      A :py:class:`xanespy.edges.KEdge` object that contains a
      description of the elemental edge being studied.
    sensitivity : float, optional
      A multiplier for the otsu value to determine the actual
      threshold.
    min_size : int, optional
      Objects below this size (in pixels) will be removed. Passing
      zero (default) will result in no effect.
    
    Returns
    -------
    mask : numpy.ndarray
      A boolean mask with the same shape as the last two dimensions of
      ``frames`` where True pixels are likely to be background
      material.
    
    """
    # Dividing the edge jump by the standard deviation provides sharp constrast
    ej = k_edge_jump(frames=frames, energies=energies, edge=edge)
    stdev = np.std(frames, axis=tuple(range(0, len(frames.shape)-2)))
    edge_ratio = ej / stdev
    # Thresholding to separate background from foreground
    img_bottom = edge_ratio.min()
    threshold = filters.threshold_otsu(edge_ratio)
    threshold = img_bottom + sensitivity * (threshold - img_bottom)
    mask = edge_ratio > threshold
    # Remove left-over background fuzz
    if min_size > 0:
        mask = morphology.opening(mask, selem=morphology.disk(min_size))
    # Invert so it removes background instead of particles
    mask = np.logical_not(mask)
    return mask


def k_edge_jump(frames: np.ndarray, energies: np.ndarray, edge):
    """Determine what the difference is between the post_edge and the
    pre_edge."""
    # Check that dimensions match
    if not frames.shape[0] == energies.shape[0]:
        msg = "First dimenions of frames and energies do not match ({} vs {})"
        raise ValueError(msg.format(frames.shape[0], energies.shape[0]))
    pre_edge = edge.pre_edge
    post_edge = edge.post_edge
    # Prepare masks for the post-edge and the pre-edge
    pre_edge_mask = np.logical_and(np.greater_equal(energies, pre_edge[0]),
                                   np.less_equal(energies, pre_edge[1]))
    if not np.any(pre_edge_mask):
        raise XanesMathError("Could not find pre-edge "
                             "{} in {}".format(pre_edge, energies))
    post_edge_mask = np.logical_and(np.greater_equal(energies, post_edge[0]),
                                    np.less_equal(energies, post_edge[1]))
    if not np.any(post_edge_mask):
        raise XanesMathError("Could not find post-edge "
                             "{} in {}".format(post_edge, energies))
        
    # Compare the post-edges and pre-edges
    mean_pre = np.mean(frames[pre_edge_mask], axis=0)
    mean_post = np.mean(frames[post_edge_mask], axis=0)
    ej = mean_post - mean_pre
    return ej


def particle_labels(frames: np.ndarray, energies: np.ndarray, edge,
                    min_distance=20):
    """Prepare a map by segmenting the images into particles.
    
    Arguments
    ---------
    frames : numpy.ndarray
      An array of images, each one at a different energy. These will
      be merged and used for segmentation.
    
    """
    # Get edge-jump mask
    mask = ~edge.mask(frames, energies=energies, min_size=0)
    # Create "edge-distance" image
    distances = ndimage.distance_transform_edt(mask)
    in_range = (distances.min(), distances.max())
    distances = exposure.rescale_intensity(distances,
                                  in_range=in_range,
                                  out_range=(0, 1))
    # Use the local distance maxima as peak centers and compute labels
    local_maxima = feature.peak_local_max(
        distances,
        indices=False,
        min_distance=min_distance,
        labels=mask.astype(np.int)
    )
    markers = measure.label(local_maxima)
    labels = morphology.watershed(-distances, markers, mask=mask)
    return labels


def direct_whitelines(spectra, energies, edge, quiet=False):
    """Takes an array of X-ray optical_depth spectra and calculates the
    positions of maximum intensities over the near-edge region.
    
    Arguments
    ---------
    spectra : np.array
      2D numpy array of optical_depth spectra where the last dimension is
      energy.
    energies : np.array
      Array of X-ray energies in electron-volts. Must be broadcastable
      to the shape of spectra.
    edge
      An XAS Edge object that describes the absorbtion edge in
      question.
    quiet : bool, optional
      Whether to suppress the progress bar, etc.
    
    Returns
    -------
    out : np.ndarray
      Array with the whiteline position of each spectrum.
    
    """
    # Broadcast energies to be same shape as spectra
    # Cut down to only those values on the edge
    edge_mask = (energies > edge.edge_range[0]) & (energies < edge.edge_range[1])
    edge_mask = np.broadcast_to(edge_mask, spectra.shape)
    # Convert energies to be same shape as spectra
    Es = np.broadcast_to(energies, spectra.shape)
    # Set values outside the mapping range to be negative infinity
    subspectra = np.copy(spectra)
    subspectra[...,~edge_mask] = -np.inf
    # Prepare a new array to hold results
    outshape = spectra.shape[:-1]
    out = np.empty(shape=outshape, dtype=energies.dtype)
    # Iterate over spectra and perform calculation

    def get_whiteline(frame_idx):
        whiteline_idx = np.argmax(subspectra[frame_idx])
        whiteline = Es[frame_idx][whiteline_idx]
        out[frame_idx] = whiteline
    indices = iter_indices(spectra, desc="Finding maxima",
                           leftover_dims=1, quiet=quiet)
    foreach(get_whiteline, indices)
    # Return results
    return out


def transform_images(data, transformations, out=None, mode='median', quiet=False):
    """Takes image data and applies the given translation matrices.
    
    It is assumed that the first dimension of `data` is the same as
    the length of `transformations`. The transformation matrices can
    be generated from translation, rotation and scale parameters via
    the :py:meth:`xanespy.xanes_math.transformation_matrics`
    function. Data will be written to `out` if given, otherwise
    returned as a new array.
    
    Arguments
    ---------
    data : np.ndarray
      Numeric array with frames to transform. Last two dimensions are
      assumed to be (row, columns).
    transformations : np.ndarray
      A numeric array shaped compatibally with `data`. The last two
      dimensions are assumed to be (3, 3) and each (3, 3) encodes a
      transformation matrix for the corresponding frame in `data`.
    out : np.ndarray, optional
      A numeric array with same shape as `data` that will hold the
      transformed data.
    mode : str, optional
      Describes how to deal with edges. See scikit-image documentation
      for options. Special value "median" (default), takes the median
      pixel intensity of that frame and uses it as the constant value.
    quiet : bool, optional
      Whether to suppress the progress bar, etc.
    
    Returns
    -------
    out : np.ndarray
      A new array with similar dimensions to `data` but with
      transformations applied and converted to float datatype.
    
    """
    logstart = time()
    # Create a new array if one is not given
    if out is None:
        out_type = np.complex if np.iscomplexobj(data) else np.float
        out = np.zeros_like(data, dtype=out_type)
    # Define a function to pass into threads
    def apply_transform(idx):
        # Get transformation parameters if given
        tmatrix = transformations[idx]
        # Prepare and execute the transformation
        transformation = transform.AffineTransform(
            matrix=tmatrix,
        )
        def do_transform(a, transformation):
            """Helper function, takes one array and transforms it."""
            realrange = (np.min(a), np.max(a))
            # Convert to float so the warp function is happy
            if not np.iscomplexobj(a):
                indata = a.astype(np.float64)
            assert realrange[0] != realrange[1], realrange
            indata = exposure.rescale_intensity(indata,
                                                in_range=realrange,
                                                out_range=(0, 1))
            # Log the anticipated transformation
            msg = "Transforming {idx} by scl={scale}; trn={trans}; rot={rot:.2f} rad"
            msg = msg.format(idx=idx, scale=transformation.scale,
                             trans=transformation.translation,
                             rot=transformation.rotation)
            log.debug(msg)
            # Perform the actual transformation
            if mode == "median":
                realmode = "constant"
                cval = np.median(indata)
            else:
                realmode = mode
                cval = 0.
            outdata = transform.warp(indata, transformation, order=3,
                                     mode=realmode, cval=cval)
            # Return to the original intensity range
            outdata = exposure.rescale_intensity(outdata,
                                                 in_range=(0, 1),
                                                 out_range=realrange)
            return outdata
        frame = data[idx]
        if np.iscomplexobj(frame):
            # Special handling for complex numbers
            j = complex(0, 1)
            out[idx] = (do_transform(np.real(frame), transformation) +
                        j * do_transform(np.imag(frame), transformation))
        else:
            # Only real numbers here
            out[idx] = do_transform(frame, transformation)
    # Loop through the images and apply each transformation
    log.debug("Starting transformations")
    indices = iter_indices(data, desc='Transforming', leftover_dims=2, quiet=quiet)
    foreach(apply_transform, indices)
    log.debug("Finished transformations in %d sec", time() - logstart)
    return out


def transformation_matrices(translations=None, rotations=None,
                            scales=None, center=(0, 0)):
    """Takes array of operations and calculates (3, 3) transformation
    matrices.
    
    This function operates by calculating an AffineTransform similar
    to that described in the scikit-image package.
    
    All three arguments (`translations`, `rotations`, and `scales`,
    should have shapes that are compatible with the frame data, though
    this is not strictly enforced for now. Rotation will necessarily
    have one less degree of freedom than translation/scale values.
    
    Example Shapes:
    
    +----------------------------+--------------+-------------+-------------+
    | Frames                     | Translations | Rotations   | Scales      |
    +============================+==============+=============+=============+
    | (10, 48, 1024, 1024)       | (10, 48, 2)  | (10, 48, 1) | (10, 48, 2) |
    +----------------------------+--------------+-------------+-------------+
    | (10, 48, 1024, 1024, 1024) | (10, 48, 3)  | (10, 48, 2) | (10, 48, 3) |
    +----------------------------+--------------+-------------+-------------+
    
    Parameters
    ----------
    translations : np.ndarray, optional
      How much to move each axis (x, y[, z]).
    
    rotations : np.ndarray, optional
      How much to rotate around the origin (0, 0) pixel.
    
    center : np.ndarray, optional
      Where to set the origin of rotation. Default is the
      first pixel (0, 0).
    
    scales : np.ndarray, optional
      How much to scale the image by in each dimension
      (x, y[, z]).
    
    Returns
    -------
    new_transforms : np.ndarray
      Resulting transformation matrices. Will have the same shape as the
      input arrays but with the last dimension replaced by (3, 3).
    
    """
    spatial_dims = 2
    assert spatial_dims == 2 # Only can handle 2D data for now
    # Get the shape of the final matrix by inspecting the inputs' shapes
    if translations is not None:
        master = translations
    elif rotations is not None:
        master = rotations
    elif scales is not None:
        master = scales
    else:
        raise ValueError("No transformations specified.")
    matrix_shape = (*master.shape[0:-1], spatial_dims+1, spatial_dims+1)
    # Prepare dummy arrays if some are missing
    if translations is None:
        translations = np.zeros((*matrix_shape[0:-2], spatial_dims))
    if rotations is None:
        rotations = np.zeros((*matrix_shape[0:-2], spatial_dims - 1))
    if scales is None:
        scales = np.ones((*matrix_shape[0:-2], spatial_dims))
    # Calculate values for transformation matrix
    #   Values taken from skimage documentations for transform.AffineTransform
    sx, sy = np.moveaxis(scales, -1, 0)
    tx, ty = np.moveaxis(translations, -1, 0)
    r, = np.moveaxis(rotations, -1, 0)
    # Calculate new eigenvectors for this transformation
    ihat = [ sx * np.cos(r), sx * np.sin(r), np.zeros_like(sx)]
    jhat = [-sy * np.sin(r), sy * np.cos(r), np.zeros_like(sx)]
    khat = [ tx,             ty,             np.ones_like(sx) ]
    # Make the eigenvectors into column vectored matrix
    new_transforms = np.array([ihat, jhat, khat]).swapaxes(0, 1)
    # Move the transformations so they're in frame order
    new_transforms = np.moveaxis(np.moveaxis(new_transforms, 0, -1), 0, -1)
    return new_transforms


def register_correlations(frames, reference, upsample_factor=10,
                          desc="Registering", quiet=False):
    """Calculate the relative translation between the reference image and
    a series of frames.
    
    This uses phase correlation through scikit-image's
    `register_translation` function.
    
    Parameters
    ----------
    frames : np.ndarray
      Array where the last two dimensions are (column, row) of images
      to be registered.
    reference : np.ndarray
      Image frame against which to align the entries in `frames`.
    upsample_factor : int, optional
      Factor controls subpixel registration via scikit-image.
    desc : str, optional
      Description for putting in the progress bar.
    quiet : bool, optional
      Whether to suppress the progress bar, etc.
    
    Returns
    -------
    translations : np.ndarray
      Array with same dimensions as 0-th axis of `frames` containing
      (x, y) translations for each frame.
    
    """
    t_shape = (*frames.shape[:-2], 2)
    translations = np.empty(shape=t_shape, dtype=np.float)
    
    def get_translation(idx):
        frm = frames[idx]
        results = feature.register_translation(reference,
                                               frm,
                                               upsample_factor=upsample_factor)
        shift, error, diffphase = results
        log.debug("Translation for frame %s = %s", str(idx), str(shift))
        # Convert (row, col) to (x, y)
        translations[idx] = (shift[1], shift[0])
    indices = iter_indices(frames, desc=desc, leftover_dims=2, quiet=quiet)
    foreach(get_translation, indices)
    # Negative in order to properly register with transform_images method
    translations = -translations
    return translations


def register_template(frames, reference, template, desc="Registering", quiet=False):
    """Calculate the relative translation between the reference image and
    a series of frames.
    
    This uses template cross correlation through scikit-image's
    `match_template` function.
    
    The `register_correlations` algorithm is simpler to use in most
    cases but sometimes results in unreasonable results; in those
    cases, this method can be more reliable to achieve a first
    approximation.
    
    Parameters
    ----------
    frames : np.ndarray
      Array where the last two dimensions are (column, row) of images
      to be registered.
    reference : np.ndarray
      Image frame against which to align the entries in `frames`.
    template : np.ndarray
      A 2D array (smaller than frames and reference) that will be
      identified in each frame and used for alignment.
    desc : str, optional
      Description for putting in the progress bar.
    quiet : bool, optional
      Whether to suppress the progress bar, etc.
    
    Returns
    -------
    translations : np.ndarray
      Array with same dimensions as 0-th axis of `frames` containing
      (x, y) translations for each frame.
    
    """
    t_shape = (*frames.shape[:-2], 2)
    translations = np.empty(shape=t_shape, dtype=np.float)
    ref_match = feature.match_template(reference, template)
    ref_center = np.unravel_index(np.argmax(ref_match), ref_match.shape)
    ref_center = np.array(ref_center)

    def get_translation(idx):
        frm = frames[idx]
        match = feature.match_template(frm, template)
        center = np.unravel_index(np.argmax(match), match.shape)
        shift = ref_center - np.array(center)
        # Convert (row, col) to (x, y)
        translations[idx] = (shift[1], shift[0])
    indices = iter_indices(frames, desc=desc, leftover_dims=2, quiet=quiet)
    foreach(get_translation, indices)
    # Negative in order to properly register with transform_images method
    translations = -translations
    return translations



# -*- coding: utf-8 -*-
#
# Copyright © 2016 Mark Wolf
#
# This file is part of Xanespy.
#
# Xanespy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Xanespy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Xanespy. If not, see <http://www.gnu.org/licenses/>.

"""Tools for accessing TXM data stored in an HDF5 file."""

import h5py
import numpy as np
import logging

from exceptions import (GroupKeyError, CreateGroupError, FrameSourceError,)

import xanes_math as xm


log = logging.getLogger(__name__)


class TXMDataset():
    """Data descriptor for accessing HDF datasets.
    
    Parameters
    ----------
    name : str
      The dataset name in the HDF file.
    context : str, optional
      Type of dataset this is: frameset, map, metadata, etc.
    dtype : np.dtype, optional
      The data-type to use when saving new data to disk. Using lower
      precision datatypes can save significant disk space.
    
    """
    def __init__(self, name, context=None, dtype=None):
        self.name = name
        self.context = context
        self.dtype = dtype
    
    def __get__(self, store, type=None):
        dataset = store.get_dataset(self.name)
        return dataset
    
    def __set__(self, store, value):
        store.replace_dataset(name=self.name, data=value,
                              context=self.context, dtype=self.dtype)
    
    def __delete__(self, store):
        del store.data_group()[self.name]


class TXMStore():
    """Wrapper around HDF5 file that stores TXM data.
    
    It has a series of descriptors and properties that return the
    corresponding HDF5 dataset object; the TXMStore().attribute.value
    pattern can be used to get pure numpy arrays directly. These
    objects should be used as a context manager to ensure that the
    file is closed, especially if using a writing mode:
    
        with TXMStore() as store:
            # Do stuff with store here
    
    Parameters
    ----------
    hdf_filename : str
      Path to the HDF file to be used.
    parent_name : str
      Name of the top-level HDF5 group.
    data_name : str
      Name of the second level HDF5 group, used for specific data
      iterations (eg. imported, aligned)
    mode : str
      Eg. 'r' for read-only, 'r+' for read-write. Passed directly to
      h5py.File constructor.
    
    """
    VERSION = 1
    _data_name = None
    
    # HDF5 Descriptors
    # ----------------
    energies = TXMDataset('energies', context='metadata')
    signals = TXMDataset('extracted_signals', context='metadata')
    original_positions = TXMDataset('original_positions', context='metadata')
    timestep_names = TXMDataset('timestep_names', context='metadata')
    pixel_sizes = TXMDataset('pixel_sizes', context='metadata')
    linear_combination_sources = TXMDataset('linear_combination_sources', context='metadata')
    
    intensities = TXMDataset('intensities', context='frameset')
    optical_depths = TXMDataset('optical_depths', context='frameset')
    references = TXMDataset('references', context='frameset')
    signal_weights = TXMDataset('extracted_signal_weights', context='frameset')
    linear_combination_parameters = TXMDataset('linear_combination_parameters', context='frameset')
    
    optical_depth_mean = TXMDataset('optical_depth_mean', context='map')
    intensity_mean = TXMDataset('intensity_mean', context='map')
    signal_map = TXMDataset('signal_map', context='map')
    edge_mask = TXMDataset('edge_mask', context='map')
    whiteline_max = TXMDataset('whiteline_max', context='map')
    whiteline_fit = TXMDataset('whiteline_fit', context='map')
    cluster_fit = TXMDataset('cluster_fit', context='map')
    particle_labels = TXMDataset('particle_labels', context='map')
    segments = TXMDataset('segments', context='map')
    linear_combination_residuals = TXMDataset('linear_combination_residuals', context='map')
    
    def __init__(self, hdf_filename: str,
                 parent_name: str, data_name=None,
                 mode='r'):
        self.hdf_filename = hdf_filename
        self._file = self.open_file(self.hdf_filename, mode=mode)
        self.parent_name = parent_name
        self.mode = mode
        # Use the latest_data_name if one isn't provided
        if data_name is None:
            data_name = self.latest_data_name
        self._data_name = data_name
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        self.close()
    
    def frameset_names(self):
        """Returns a list of all the valid map representations."""
        names = []
        for key in self.data_group().keys():
            # Check if it's a "map" or not
            grp = self.data_group()
            context = grp[key].attrs.get('context', '')
            if context == 'frameset':
                names.append(key)
        return names        
    
    def map_names(self):
        """Returns a list of all the valid map representations."""
        names = []
        for key in self.data_group().keys():
            # Check if it's a "map" or not
            if self.data_group()[key].attrs['context'] == 'map':
                names.append(key)
        return names
    
    def open_file(self, filename, mode):
        return h5py.File(filename, mode=mode)
    
    def close(self):
        self._file.close()
    
    @property
    def data_name(self):
        return self._data_name
    
    @data_name.setter
    def data_name(self, val):
        if val not in self.parent_group().keys():
            msg = "Group {} does not exists. Run TXMStore.fork_data_group('{}') first"
            raise CreateGroupError(msg.format(val, val))
        self._data_name = val
    
    def data_tree(self):
        """Create a tree of the possible groups this store could access. The
        first level is samples, then data_groups (ie. same sample but
        different analysis status), then representations.
        """
        # Define a recursive function to walk the groups in the file
        def walk_groups(parent, level):
            # Recurse into children
            keys = getattr(parent, 'keys', lambda: [])
            datas = []
            for key in keys():
                # Check for whether this object is a frameset or a map
                node = parent[key]
                grpdata = {
                    'name': key,
                    'path': node.name,
                    'context': node.attrs.get('context', None),
                    'ndim': getattr(node, 'ndim', 0),
                    'level': level,
                    # Resurse
                    'children': walk_groups(parent=node, level=level+1),
                }
                datas.append(grpdata)
            return datas
        # Start the recursion at the top
        tree = walk_groups(self._file, level=0)
        return tree
    
    def fork_data_group(self, dest, src=None):
        """Turn on different active data group for this store. This method
        deletes the existing group and copies symlinks from the
        current one.
        
        """
        # Switch to the group given by `src`
        if src is not None:
            self.data_name = src
        # Check that the current and target groups are not the same
        if dest == self.data_name:
            log.critical('Refusing to fork group "%s" to itself', dest)
            msg = 'Refusing to fork myself to myself ({})'.format(dest)
            raise CreateGroupError(msg)
        log.info('Forking data group "%s" to "%s"', src, dest)
        # Delete the old group and overwrite it
        parent = self.parent_group()
        if dest in parent.keys():
            del parent[dest]
        # Copy the old one with symlinks
        new_group = parent.copy(self.data_group(), dest, shallow=True)
        self.latest_data_name = dest
        self.data_name = dest
        return new_group
    
    @property
    def latest_data_name(self):
        name = self.parent_group().attrs['latest_data_name']
        return name
    
    @latest_data_name.setter
    def latest_data_name(self, val):
        self.parent_group().attrs['latest_data_name'] = val
    
    def validate_parent_group(self, name):
        """Retrieve the real parent group name for a possible parent_group.
        
        If ``name`` is None and only one group exists in the file,
        then that group name will be returned. If ``name`` is in the
        file, then ``name`` will be returned. If ``name`` is not in
        the file, a GroupKeyError will be raised.
        """
        if name is None and len(self._file.keys()) == 1:
            new_name = list(self._file.keys())[0]
        elif name not in self._file.keys():
            raise GroupKeyError("Cannot load parent group '{group}'. "
                                "Valid choices are {choices}."
                                "".format(group=name, choices=list(self._file.keys())))
        else:
            new_name = name
        return new_name
    
    def parent_group(self):
    
        """Retrieve the top-level HDF5 group object for this file and
        groupname."""
        try:
            group = self._file[self.parent_name]
        except (TypeError, KeyError):
            # Invalid group name, throw an exception
            msg = 'Cannot load parent group "{group}". Valid choices are {choices}'
            choices = list(self._file.keys())
            msg = msg.format(group=self.parent_name, choices=choices)
            raise GroupKeyError(msg) from None
        return group
    
    def data_group(self):
        """Retrieve the currently active second-level HDF5 group object for
        this file and groupname. Ex. "imported" or "aligned_frames".
        """
        return self.parent_group()[self.data_name]
    
    def replace_dataset(self, name, data, context=None, attrs={},
                        compression=None, *args, **kwargs):
        """Wrapper for h5py.create_dataset that removes the existing dataset
        if it exists.
        
        Parameters
        ----------
        name : str
          HDF5 groupname name to give this dataset.
        data : np.ndarray
          Numpy array of data to be saved.
        context : str, optional
          Specifies what kind of data is stored. Eg. "frameset",
          "metadata", "map".
        attrs : dict, optional
          Dictionary containing HDF5 metadata attributes to be set on
          the resulting dataset.
        *args
          Arguments to pass to h5py's ``create_dataset`` method.
        **kwargs
          Keyword arguments to pass to h5py's ``create_dataset`` method.
        """
        # Remove the existing dataset if possible
        try:
            del self.data_group()[name]
        except KeyError:
            pass
        # Perform the actual group creation
        ds = self.data_group().create_dataset(name=name, data=data,
                                              compression=compression,
                                              *args, **kwargs)
        # Set metadata attributes
        if context is not None:
            ds.attrs['context'] = context
        for key, val in attrs.items():
            ds.attrs[key] = val
    
    def get_dataset(self, name):
        """Attempt to open the requested dataset.
        
        Returns
        -------
        data : hyp5.Dataset
          An open HDF5 dataset
        
        Raises
        ------
        exceptions.GroupKeyError
          If the dataset does not exist in the file.
        """
        # Check for some bad dataset names
        if name is None:
            msg = "dataset `None` not found in file '{}'"
            msg = msg.format(self.hdf_filename)
            raise GroupKeyError(msg)
        elif name not in self.data_group().keys():
            msg = "dataset '{}' not found in group '{}' file '{}'"
            msg = msg.format(name, self.data_group().name, self.hdf_filename)
            raise GroupKeyError(msg)
        else:
            data = self.data_group()[name]
        return data
    
    def has_dataset(self, name):
        """Return a boolean indicated whether this dataset exists in the HDF
        file.
        
        """
        try:
            result = name in self.data_group().keys()
        except TypeError:
            result = False
        return result
    
    def frame_source(self, name):
        """Get the name of the frames that went into creating a map."""
        attrs = getattr(self.get_dataset(name), 'attrs', {})
        source = attrs.get('frame_source', name)
        return source
    
    def get_frames(self, name):
        """Return the source frame data for the given data name.
        
        This is similar to ``get_dataset`` except that if the
        data are a map, then get the frames that went into making it.
        
        Parameters
        ----------
        name : str
          The dataset name for which to retrieve frames.
        
        Returns
        -------
        dataset : h5py.Dataset
          The requested frameset. If the dataset is actually a map, as
          determined by the "context" attribute, then the related
          frame source attribute will be retrieved.
        
        """
        dataset = self.get_dataset(name)
        # If it's a map, then return the source frames instead
        if dataset.attrs.get('context', None) == 'map':
            try:
                dataset = self.get_dataset(dataset.attrs['frame_source'])
            except KeyError:
                source_desc = dataset.attrs.get('frame_source', 'None')
                # raise FrameSourceError(
                log.warning(
                    "Invalid frame source {} specified for group {}"
                    "".format(source_desc, self.data_name))
                dataset = self.get_dataset('optical_depths')
        return dataset
    
    @property
    def relative_positions(self):
    
        """(x, y, z) position values for each frame."""
        return self.data_group()['relative_positions']
    
    @relative_positions.setter
    def relative_positions(self, val):
        self.replace_dataset('relative_positions', val, context='metadata')
        self.data_group()['relative_positions'].attrs['order'] = "(x, y, z)"
   
    @property
    def pixel_unit(self):
        return self.data_group()['pixel_sizes'].attrs['unit']
    
    @pixel_unit.setter
    def pixel_unit(self, val):
        self.data_group()['pixel_sizes'].attrs['unit'] = val
    
    @property
    def signal_method(self):
        """String describing how the previously extracted signals were
        calculated.
        """
        return self.data_group()['extracted_signals'].attrs['method']
    
    @signal_method.setter
    def signal_method(self, val):
        self.data_group()['extracted_signals'].attrs['method'] = val
    
    @property
    def timestamps(self):
        return self.data_group()['timestamps']
    
    @timestamps.setter
    def timestamps(self, val):
        # S32 is the 32-character ACSII string type for numpy
        val = np.array(val, dtype="S32")
        self.replace_dataset('timestamps', val, dtype="S32", context='metadata',
                             attrs={'timezone', "UTC"})
    
    @property
    def filenames(self):
        return self.data_group()['filenames']
    
    @filenames.setter
    def filenames(self, val):
        # S100 is the 100-character ACSII string type for numpy
        val = np.array(val, dtype="S100")
        self.replace_dataset('filenames', val, dtype="S100", context='metadata')
    
    @property
    def fit_parameters(self):
        return self.get_dataset('fit_parameters')
    
    @fit_parameters.setter
    def fit_parameters(self, val):
        attrs = {
            'parameters': str(xm.kedge_params),
        }
        return self.replace_dataset('fit_parameters', val,
                                    attrs=attrs, context="metadata",
                                    dtype=np.float64)



#!/usr/bin/env python

import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name="xanespy",
      version=read("VERSION"),
      description="Tools for analyzing X-ray absorbance spectroscopy data",
      long_description=read('README.rst'),
      long_description_content_type='text/x-rst',
      author="Mark Wolfman",
      author_email="canismarko@gmail.com",
      url="https://github.com/canismarko/xanespy",
      keywords="XANES X-ray microscopy operando",
      install_requires=[
          'pytz>=2013b', 'h5py', 'pandas', 'olefile', 'matplotlib', 'scikit-image',
          'scikit-learn', 'numpy',
      ],
      packages=['xanespy',],
      package_data={
          'xanespy': ['qt_map_window.ui', 'qt_frame_window.ui']
      },
      entry_points={
          'gui_scripts': [
              'xanespy-viewer = xanespy.xanes_viewer:main'
          ]
      },
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Natural Language :: English',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 3.5',
          'Topic :: Scientific/Engineering :: Chemistry',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Scientific/Engineering :: Visualization',
      ]
)



# -*- coding: utf-8 -*-
#
# Copyright © 2016 Mark Wolf
#
# This file is part of Xanespy.
#
# Xanespy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Xanespy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Xanespy.  If not, see <http://www.gnu.org/licenses/>.

"""A collection of classes and functions that arent' specific to any
one type of measurement. Also, it defines some namedtuples for
describing coordinates.
"""

from collections import namedtuple
import multiprocessing
import threading
import sys
import logging
import math

import numpy as np
import h5py

# Check if the tqdm package is installed for showing progress bars
try:
    from tqdm import tqdm, tqdm_notebook
except (ImportError):
    HAS_TQDM = False
else:
    HAS_TQDM = True


position = namedtuple('position', ('x', 'y', 'z'))
xycoord = namedtuple('xycoord', ('x', 'y'))
Pixel = namedtuple('Pixel', ('vertical', 'horizontal'))
shape = namedtuple('shape', ('rows', 'columns'))
Extent = namedtuple('extent', ('left', 'right', 'bottom', 'top'))


CPU_COUNT = multiprocessing.cpu_count()

def foreach(f, l, threads=CPU_COUNT, return_=False):
    """
    Apply f to each element of l, in parallel
    """
    if threads>1:
        iteratorlock = threading.Lock()
        exceptions = []
        if return_:
            n = 0
            d = {}
            i = zip(count(),l.__iter__())
        else:
            i = l.__iter__()
        
        
        def runall():
            while True:
                iteratorlock.acquire()
                try:
                    try:
                        if exceptions:
                            return
                        v = next(i)
                    finally:
                        iteratorlock.release()
                except StopIteration:
                    return
                try:
                    if return_:
                        n,x = v
                        d[n] = f(x)
                    else:
                        f(v)
                except:
                    e = sys.exc_info()
                    iteratorlock.acquire()
                    try:
                        exceptions.append(e)
                    finally:
                        iteratorlock.release()
        threadlist = [threading.Thread(target=runall) for j in range(threads)]
        for t in threadlist:
            t.start()
        for t in threadlist:
            t.join()
        if exceptions:
            a, b, c = exceptions[0]
            exception = a(b)
            exception.with_traceback(c)
            raise exception
        if return_:
            r = list(d.items())
            r.sort()
            return [v for (n,v) in r]
    else:
        if return_:
            return [f(v) for v in l]
        else:
            for v in l:
                f(v)
            return


def parallel_map(f,l,threads=CPU_COUNT):
    return foreach(f,l,threads=threads,return_=True)


def xy_to_pixel(xy, extent, shape):
    """Take an xy location on an image and convert it to a pixel location
    suitable for numpy indexing."""
    ratio_x = (xy.x - extent.left) / (extent.right - extent.left)
    pixel_h = int(round(ratio_x * shape[1]))
    ratio_y = (extent.top - xy.y) / (extent.top - extent.bottom)
    # (1 - ratio) for y because images are top indexed
    pixel_v = int(round(ratio_y * shape[0]))
    # Very right and top edges get set to the last pixel
    pixel_v = min(pixel_v, shape[0] - 1)
    pixel_h = min(pixel_h, shape[1] - 1)
    return Pixel(vertical=pixel_v, horizontal=pixel_h)


def pixel_to_xy(pixel, extent, shape):
    """Take an xy location on an image and convert it to a pixel location
    suitable for numpy indexing."""
    ratio_h = ((pixel.horizontal+0.5) / shape[1])
    x = extent.left + ratio_h * (extent.right - extent.left)
    ratio_v = ((pixel.vertical+0.5) / shape[0])
    y = extent.top - ratio_v * (extent.top - extent.bottom)
    return xycoord(x=x, y=y)


def broadcast_reverse(array, shape, *args, **kwargs):
    """Take the array and extends it as much as possible to match
    `shape`. Similar to numpy's broadcast_to function, but starts with
    the most significant axis. For example, if `array` has shape (7,
    29), it can be broadcast to (7, 29, 1024, 1024).
    
    """
    def reverse_axes(array):
        ndims = array.ndim
        for n in range(int(ndims/2)):
            array = np.swapaxes(array, n, ndims - 1 - n)
        return array
    # Convert the array to its reverse representation
    rev_array = reverse_axes(array)
    rev_shape = shape[::-1]
    # Broadcast the array
    new_array = np.broadcast_to(rev_array, rev_shape, *args, **kwargs)
    # Convert the array back to its original representation
    new_array = reverse_axes(new_array)
    return new_array


def get_component(data, name):
    """If complex, turn to given component, otherwise return original data.
    
    Arguments
    ---------
    data : 
      Numerical (presumably complex-valued) data to be reduced to a
      component.
    name : str
      One of ('modulus', 'phase', 'real', 'imag')
    
    Returns
    -------
    data : 
      Input data converted to requested component.
    """
    if np.iscomplexobj(data):
        # Sort out complex components
        if name == "modulus":
            data = np.abs(data)
        elif name == "phase":
            data = np.angle(data)
        elif name == "real":
            data = np.real(data)
        elif name == "imag":
            data = np.imag(data)
        else:
            raise ValueError("Unrecognized component {}, choices are "
                             "'modulus', 'phase', 'real', 'imag'".format(name))
    else:
        # Real data so return the real part
        data = np.real(data)
    return data


def is_kernel():
    """Detect whether or not we're running inside an IPython kernel. NB:
    This does not distinguish between eg IPython notebook and IPython
    QtConsole.
    """
    if 'IPython' not in sys.modules:
        # IPython hasn't been imported, definitely not
        return False
    from IPython import get_ipython
    # check for `kernel` attribute on the IPython instance
    return getattr(get_ipython(), 'kernel', None) is not None


def prog(iterable=None, leave=None, dynamic_ncols=True, smoothing=0.01, *args, **kwargs):
    """A progress bar for displaying how many iterations have been
    completed.
    
    This is mostly just a wrapper around the tqdm library. args and
    kwargs are passed directly to either tqdm.tqdm or
    tqdm.tqdm_notebook. This function also takes into account the
    value of ``USE_PROG`` defined in this module. If tqdm is no
    installed, then calls to prog will just return the iterable again.
    
    Parameters
    ----------
    iterable
      Iterable to decorate with a progressbar. See ``tqdm.tqdm``
      documentation for more details.
    leave : bool, optional
      Whether to leave the progress bar in the stream after it's
      completed. If omitted, will depend on terminal used.
    dynamic_ncols : bool, optional
      Whether to adapt dynamically to the width of the environment.
    args
      Positional arguments passed directly to ``tqdm`` or ``tqdm_notebook``.
    kwargs
      Keyword arguments passed directly to ``tqdm`` or ``tqdm_notebook``.
    
    """
    if not HAS_TQDM:
        # No progress bar will be displayed
        prog_iter = iterable
    elif is_kernel():
        # Use the IPython widgets version of tqdm
        leave = True if leave is None else leave
        prog_iter = tqdm(iterable, leave=leave,
                         dynamic_ncols=dynamic_ncols,
                         smoothing=smoothing, *args, **kwargs)
    else:
        # Use the text-version of tqdm
        leave = False if leave is None else leave
        prog_iter = tqdm(iterable, leave=leave,
                         dynamic_ncols=dynamic_ncols,
                         smoothing=smoothing, *args, **kwargs)
    return prog_iter
